import pytest

from extradict import FallbackNormalizedDict
from extradict import NormalizedDict


def test_normalized_works():
    a = NormalizedDict(maçã=1)
    assert a["maçã"] == 1
    assert a["maca"] == 1


def test_normalized_doesnot_strip_numbers():
    a = NormalizedDict(maçã123=1)
    assert a["maca123"]
    with pytest.raises(KeyError):
        a["maca"]
        a["maca567"]


def test_fallbacknormalized_works():
    a = FallbackNormalizedDict(maçã=1)
    assert a["maçã"] == 1
    assert a["maca"] == 1
    a["maca"] = 2
    assert a["maçã"] == 1
    assert a["maca"] == 2


def test_normalized_removes_puntuation():
    a = NormalizedDict()
    a["merry, cHristmas??!"] = 2
    assert a["Merry Christmas"]


def test_fallbacknormalized_removes_puntuation():
    a = FallbackNormalizedDict()
    a["merry, cHristmas??!"] = 2
    assert a["Merry Christmas"]


def test_fallbacknormalized_deleting():
    a = FallbackNormalizedDict(maçã=1)
    with pytest.raises(KeyError):
        del a["maca"]
    del a["maçã"]
    assert "maçã" not in a


def test_fallbacknormalized_get_multi():
    a = FallbackNormalizedDict(maçã=1, maca=2)
    assert a["maca"] == 2
    assert a["maçã"] == 1
    assert a.get_multi("maca") == [1, 2]


# strip_replacer = lambda text: re.sub(r"\W", "", text)
# unicode_normalizer = lambda text: unicodedata.normalize("NFKD", text)
# case_normalizer = str.lower


def test_node_strip_replacer():
    from extradict.normalized_dict import strip_replacer

    assert strip_replacer("-1a ") == "1a"
    assert strip_replacer("-1aµ ") == "1aµ"
    assert strip_replacer("a") == "a"
    assert strip_replacer("á") != "a"
    assert strip_replacer("A") != "a"


def test_node_unicode_normalizer():
    from extradict.normalized_dict import unicode_normalizer

    assert unicode_normalizer("á") == "\u0061\u0301"


def test_node_case_normalizer():
    from extradict.normalized_dict import case_normalizer

    assert case_normalizer("PyThOn") == "python"


def test_normalized_dict_can_be_customized():
    import re

    a = NormalizedDict()
    a.pipeline = a.pipeline.copy()
    a.pipeline[1] = lambda text: re.sub("[^A-Za-z]", "", text)
    a["maçã123"] = 1
    assert a["567maca"] == 1



from collections.abc import Mapping


class TransformKeyDict(Mapping):
    """Special mapping that algoritmicaly modifies the key passed in.

    Change the "transform" attribute after instantiating or sublassing.
    The main purpose of presenting this is to have a mapping interface,
    so that it can be composed with other dicionary types.

    """

    transform = lambda s, key: key  # noqa

    def __getitem__(self, key):
        return self.transform(key)

    def __len__(self):
        raise ValueError("Echo dict has no lenght")

    def __iter__(self):
        raise NotImplementedError


EchoDict = TransformKeyDict()
# This can actually be used as a singleton.
# common recipe:
# d = collections.ChainMap(other_dict, EchoDict)



from extradict import (
    VersionDict,
    OrderedVersionDict,
    FallbackNormalizedDict,
    NormalizedDict,
    BijectiveDict,
    TreeDict,
    BlobTextDict,
)
from extradict.nested_data import _NestedDict

import pytest


"""
Tests for normal mapping usage - as in plain get/set elements, etc
"""


def all_mappings(func):
    return pytest.mark.parametrize(
        "cls",
        [
            dict,
            VersionDict,
            OrderedVersionDict,
            FallbackNormalizedDict,
            NormalizedDict,
            BijectiveDict,
            TreeDict,
            _NestedDict,
            BlobTextDict,
        ],
    )(func)


@all_mappings
def test_instantiates_empty(cls):
    inst = cls()
    assert not inst
    assert type(inst) is cls


@all_mappings
def test_creates_single_key_value_pair(cls):
    inst = cls()
    data = (
        "pair"
        if not issubclass(cls, BlobTextDict)
        else {
            "pair",
        }
    )
    inst["key"] = data
    assert (
        inst["key"] == data
        if not issubclass(cls, BlobTextDict)
        else set(inst["key"]) == data
    )
    assert "key" in inst


@all_mappings
def test_creates_several_key_value_pairs(cls):
    inst = cls()
    for i, j in zip(range(0, 10), range(10, 20)):
        i = str(i)
        inst[i] = (
            j
            if not issubclass(cls, BlobTextDict)
            else {
                str(j),
            }
        )

    for i, j in zip(range(0, 10), range(10, 20)):
        i = str(i)
        if not issubclass(cls, BlobTextDict):
            assert inst[i] == j
        else:
            assert str(j) in inst[i]


@all_mappings
def test_del_key_value_pair(cls):
    # if cls in (FallbackNormalizedDict,):
    # return
    inst = cls()
    inst["key"] = (
        "pair"
        if not issubclass(cls, BlobTextDict)
        else {
            "pair",
        }
    )
    del inst["key"]
    assert not "key" in inst
    assert not inst


@all_mappings
def test_pop(cls):
    if cls in (FallbackNormalizedDict,):
        return
    inst = cls()
    inst["key"] = (
        "pair"
        if not issubclass(cls, BlobTextDict)
        else {
            "pair",
        }
    )
    inst.pop("key")
    assert not "key" in inst
    assert not inst


@all_mappings
def test_dict_round_trip(cls):
    # 'NormalizedDict' would not round trip strings that would be have a different normal form
    # but that is not the  object of this test
    if cls in (BijectiveDict, BlobTextDict):
        return
    test = {"a": 1, "b": 2, "c": 3, "d": 4}
    inst = cls(test.copy())

    assert dict(inst) == test



import pytest

from extradict.blobdict import BlobTextDict, _BlobSets


def test_blobsets_works():
    a = _BlobSets()
    offset = a.add_new(["a"])
    assert "a" in a.get(offset)


def test_blobsets_works2():
    a = _BlobSets()
    offset = a.add_new(["a", "b"])
    assert "a" in a.get(offset)
    assert "b" in a.get(offset)


def test_blobsets_works3():
    a = _BlobSets()
    sets = [["a", "b"], ["c", "d"]]
    offset1 = a.add_new(sets[0])
    offset2 = a.add_new(sets[1])
    for offset, current_set in zip((offset1, offset2), sets):
        assert all(x in a.get(offset) for x in current_set)


def test_blobsets_works_multilength_string():
    a = _BlobSets()
    offset = a.add_new(["maçã", "banana"])
    assert "maçã" in a.get(offset)
    assert "banana" in a.get(offset)
    assert "ma" not in a.get(offset)


def test_blobsets_rejects_strings_with_null():
    a = _BlobSets()
    with pytest.raises(ValueError):
        a.add_new(["ab\x00c"])


def test_blobsets_add_new_rejects_pure_string_sequences():
    a = _BlobSets()
    with pytest.raises(TypeError):
        a.add_new("abc")


def test_blobsliceset_discard():
    a = _BlobSets()
    offset = a.add_new(["maçã", "banana"])
    a.get(offset).discard("maçã")
    assert "maçã" not in a.get(offset)
    assert "banana" in a.get(offset)
    a.get(offset).discard("banana")
    assert "banana" not in a.get(offset)


def test_blobsliceset_add_small():
    # new addition does not require allocation
    a = _BlobSets()
    offset = a.add_new(["a", "b", "c", "d"])
    a.get(offset).add("e")
    assert all(letter in a.get(offset) for letter in "abcde")


def test_blobsliceset_add_large():
    # new addition does require re-allocation
    a = _BlobSets()
    text = "abcefghijklmnopqrstuvwxyz"
    offset = a.add_new(["a"])
    b = a.get(offset)
    for letter in text[1:]:
        b.add(letter)
    assert all(letter in a.get(offset) for letter in text)
    assert a._final_offset(offset) != offset


def test_blob_text_dict_stores_character_in_list():
    a = BlobTextDict()
    a["test"] = ["a"]
    assert "a" in a["test"]


def test_blob_text_dict_has_default_sets():
    a = BlobTextDict()
    a["test"].add("a")
    assert "a" in a["test"]



import pytest

from extradict import Grouper


def test_grouper_works():
    x = Grouper(range(10))
    assert all(len(list(x[k])) == 1 for k in range(10))


def test_grouper_groups():
    x = Grouper(range(100), key=lambda x: x // 10)
    assert all(len(list(x[k])) == 10 for k in range(10))


def test_grouper_call_returns_dict_with_lists():
    x = Grouper(range(100), key=lambda x: x // 10)()
    assert x == {k: list(range(k * 10, k * 10 + 10)) for k in range(10)}


def test_grouper_call_keyhint():
    x = Grouper(range(5))(keyhint=(5, 6, 7, 8))
    assert x == {**{k: [k] for k in range(5)}, **{k: [] for k in range(5, 9)}}


def test_grouper_yields_key_error_on_unmatched_key():
    x = Grouper(range(10))
    x[0]
    with pytest.raises(KeyError):
        x[10]



from collections.abc import Mapping, Sequence
from copy import deepcopy

from extradict import NestedData
from extradict.nested_data import _NestedDict, _NestedList

import pytest


@pytest.mark.parametrize(["initial"], [({},), ([],)])
def test_can_build_nested_data_from_empty_object(initial):
    nested = NestedData(initial)
    assert isinstance(nested.data, type(initial))


def test_nested_data_composite_key():
    a = NestedData({"person.address.city": "São Paulo"})
    assert a["person.address.city"] == "São Paulo"
    assert "person" in a
    assert "address" in a["person"]
    assert "city" in a["person"]["address"]
    assert a["person"]["address"]["city"] == "São Paulo"
    assert a["person.address"]["city"] == "São Paulo"
    assert a["person"]["address.city"] == "São Paulo"
    assert "city" not in a["person"]
    assert "address" not in a


def test_nested_data_creates_mappings():
    a = NestedData({"person.address.city": "São Paulo"})
    assert isinstance(a, NestedData)
    assert isinstance(a, Mapping)


def test_nested_data_can_be_assigned():
    address = {"city": "São Paulo", "street": "Av. Paulista", "number": 37}
    a = NestedData({"person.address": {}})
    a["person.address"] = address
    assert a["person.address.street"] == "Av. Paulista"
    assert a["person.address.city"] == "São Paulo"
    assert a["person.address.number"] == 37


def test_nested_data_new_data_is_replaced_on_assignment():
    address = {"city": "São Paulo", "street": "Av. Paulista"}
    extra = {"number": 37, "cep": "01311-902"}
    a = NestedData({"person.address": {}})
    a["person.address"] = address
    a["person.address"] = extra
    assert "street" not in a["person.address.street"]
    assert "city" not in a["person.address.city"]
    assert a["person.address.number"] == 37
    assert a["person.address.cep"] == "01311-902"


def test_nested_data_new_data_is_merged():
    address = {"city": "São Paulo", "street": "Av. Paulista"}
    extra = {"number": 37, "cep": "01311-902"}
    a = NestedData({"person.address": {}})
    a["person.address"] = address
    a["person.address"].merge(extra)
    assert a["person.address.street"] == "Av. Paulista"
    assert a["person.address.city"] == "São Paulo"
    assert a["person.address.number"] == 37
    assert a["person.address.cep"] == "01311-902"


def test_nested_data_new_data_is_merged_with_path():
    address = {"city": "São Paulo", "street": "Av. Paulista"}
    extra = {"number": 37, "cep": "01311-902"}
    a = NestedData({"person.address": {}})
    a["person.address"] = address
    a.merge(extra, path="person.address")
    assert a["person.address.street"] == "Av. Paulista"
    assert a["person.address.city"] == "São Paulo"
    assert a["person.address.number"] == 37
    assert a["person.address.cep"] == "01311-902"


def test_nested_data_distinct_blocks_can_be_assigned_and_contains_works_with_path():
    address = {"city": "São Paulo", "street": "Av. Paulista"}
    contacts = {"email": "tarsila@example.com"}
    extra = {
        "person": {
            "contacts": {"phone": "+55 11 5555-1234"},
            "address": {"number": 37, "cep": "01311-902"},
        }
    }
    a = NestedData({"person": {}})
    a["person.address"] = address
    a["person.contacts"] = contacts

    assert "person.address.street" in a
    assert "person.contacts.email" in a
    assert a["person.contacts.email"] == "tarsila@example.com"


def test_nested_data_new_data_is_deeply_merged():
    address = {"city": "São Paulo", "street": "Av. Paulista"}
    contacts = {"email": "tarsila@example.com"}
    extra = {
        "person": {
            "contacts": {"phone": "+55 11 5555-1234"},
            "address": {"number": 37, "cep": "01311-902"},
        }
    }
    a = NestedData({"person": {}})
    a["person.address"] = address
    a["person.contacts"] = contacts

    a.merge(extra)

    assert "person.contacts.email" in a
    assert a["person.contacts.email"] == "tarsila@example.com"
    assert "person.contacts.phone" in a
    assert a["person.contacts.phone"] == "+55 11 5555-1234"
    assert a["person.address.street"] == "Av. Paulista"
    assert a["person.address.city"] == "São Paulo"
    assert a["person.address.number"] == 37
    assert a["person.address.cep"] == "01311-902"


def test_inner_data_structures_present_themselves_as_nested_data_instances():
    assert isinstance(_NestedDict(), NestedData)
    assert isinstance(_NestedList(), NestedData)
    assert isinstance(NestedData(), NestedData)


def test_nested_data_instances_compare_equal_to_unwrapped_data():
    assert NestedData({"a": 1}) == {"a": 1}
    assert NestedData([1, 2, 3]) == [1, 2, 3]


def test_retrieving_subtree_returns_nested_data_instances():
    a = NestedData({"a": [1, 2, 3, {"b": 2}]})
    assert isinstance(a["a"], NestedData)
    assert isinstance(a["a.3"], NestedData)
    assert isinstance(a["a"][3], NestedData)
    assert isinstance(a["a.2"], int)


def test_nested_data_can_delete_deep_elements():
    a = NestedData({"person.address.city": "São Paulo"})
    a["person.address.street"] = "Av. Paulista"

    del a["person.address.city"]
    assert "person.address.street" in a
    del a["person.address.street"]

    assert a["person.address"] == {}
    del a["person.address"]

    with pytest.raises(KeyError):
        a["person.address"]


#########################


def test_nested_data_composite_key_creates_sequences_with_numeric_indexes():
    a = NestedData({"0": "São Paulo", "1": "Rio de Janeiro"})
    assert a.data == ["São Paulo", "Rio de Janeiro"]
    assert isinstance(a, NestedData)
    assert isinstance(a, Sequence)


def test_nested_data_composite_key_creates_sequences_with_numeric_indexes_as_ints():
    a = NestedData({0: "São Paulo", 1: "Rio de Janeiro"})
    assert a.data == ["São Paulo", "Rio de Janeiro"]
    assert isinstance(a, NestedData)
    assert isinstance(a, Sequence)


def test_nested_data_creates_sequences_from_sequences():
    a = NestedData(["São Paulo", "Rio de Janeiro"])
    assert a.data == ["São Paulo", "Rio de Janeiro"]
    assert isinstance(a, Sequence)


def test_nested_data_creates_sequences_from_sets():
    a = NestedData({"São Paulo", "Rio de Janeiro"})
    assert isinstance(a, Sequence)
    assert set(a.data) == set(["São Paulo", "Rio de Janeiro"])


def test_nested_data_numeric_indexes_require_all_keys():
    with pytest.raises(ValueError):
        a = NestedData({"0": "São Paulo", "2": "Rio de Janeiro"})
    a = NestedData({"0": "São Paulo", "2": "Rio de Janeiro"}, default=None)
    assert a.data == ["São Paulo", None, "Rio de Janeiro"]
    assert isinstance(a, Sequence)


def test_nested_data_composite_key_creates_sequences_with_nested_numeric_indexes():
    a = NestedData({"cities": {"0": "São Paulo", "1": "Rio de Janeiro"}})
    assert a["cities"] == ["São Paulo", "Rio de Janeiro"]
    assert isinstance(a, NestedData)
    assert isinstance(a, Mapping)
    assert isinstance(a["cities"], Sequence)


def test_nest_data_sequence_atribution_with_star_index_changes_all_sub_values():

    a = NestedData(
        {
            "data": [
                {"detail": {"state": "M"}},
                {"detail": {"state": "N"}},
                {"detail": {"state": "O"}},
                {"detail": {"state": "D"}},
                {"detail": {"state": "X"}},
            ]
        }
    )

    a["data.*.detail.state"] = "A"

    for item in a["data"]:
        assert isinstance(item, NestedData)
        assert item["detail.state"] == "A"


@pytest.mark.parametrize(
    ["path", "target"],
    [
        ("data.*", {"detail": {"state": "A"}}),
        ("data.*.detail", {"state": "A"}),
        ("data.*.detail.state", "A"),
    ],
)
def test_nest_data_sequence_merge_with_star_index_changes_all_sub_values(path, target):
    a = NestedData(
        {
            "data": [
                {"detail": {"state": "M", "other": 0}},
                {"detail": {"state": "N", "other": 1}},
                {"detail": {"state": "O", "other": 2}},
                {"detail": {"state": "D", "other": 3}},
                {"detail": {"state": "X", "other": 4}},
            ]
        }
    )
    a.merge(target, path)

    for i, item in enumerate(a["data"]):
        assert isinstance(item, NestedData)
        assert item["detail.state"] == "A"
        assert item["detail.other"] == i


def test_nested_data_sequence_can_shallow_merge_ifitem_not_container():
    a = NestedData([0, 1, 2])
    a.merge(3, 0)
    assert a[0] == 3


def test_nested_data_sequence_cant_deep_merge_ifitem_not_container():
    a = NestedData([0, 1, 2])
    with pytest.raises(IndexError):
        a.merge(1, "0.b")


@pytest.mark.parametrize(["path"], [(0,), ("0",)])
def test_nested_data_sequence_can_merge_ifitem_dict(path):
    a = NestedData([{}, 1, 2])
    a.merge({"b": 1}, path)
    assert a["0.b"] == 1


def test_nested_data_sequence_can_merge_ifitem_list():
    a = NestedData([[{}], 1, 2])
    a.merge({"b": 1}, "0.0.a")
    assert a["0.0.a.b"] == 1


def test_nested_data_sequence_works_with_str_index():
    a = NestedData([10, 20, 30])
    assert a["0"] == 10


def test_nested_data_sequence_works_with_int_index():
    a = NestedData([10, 20, 30])
    assert a[0] == 10


def test_nested_data_sequence_append_root():
    a = NestedData([10, 20, 30])
    a.append(40)
    assert a[3] == 40


def test_nested_data_sequence_append_l2():
    a = NestedData({"b": [10, 20, 30]})
    a["b"].append(40)
    assert a["b.3"] == 40


# we dont want this behavior. Assign a list to the parent key instead.
# def test_nested_data_new_key_ending_in_int_creates_new_list():
# a = NestedData()
# a["a.0"] = 23
# assert a["a"].data == [23]

# def test_nested_data_new_key_containing_int_creates_new_list():
# a = NestedData()
# a["a.0.b"] = 23
# assert a["a"].data == [{"b": 23}]


def test_sequence_with_asterisk_in_index_yields_sequence_with_all_leaves():
    x = NestedData({"a": [{"b": i} for i in range(5)]})
    assert list(range(5)) == [y["b"] for y in x["a.*"]]


def test_sequence_with_asterisk_in_index_items_are_unwrapped():
    x = NestedData({"a": [{"b": i} for i in range(5)]})
    assert not isinstance(x["a.*"].data[0], NestedData)
    assert isinstance(x["a.*"].data[0], dict)


def test_sequence_root_can_be_merged():
    x = NestedData(
        [
            {"color": "red", "value": "#f00"},
            {"color": "green", "value": "#0f0"},
            {"color": "blue", "value": "#00f"},
        ]
    )
    y = NestedData([{"realm": "earth"} for _ in range(len(x))])
    z = [
        {"color": "red", "value": "#f00", "realm": "earth"},
        {"color": "green", "value": "#0f0", "realm": "earth"},
        {"color": "blue", "value": "#00f", "realm": "earth"},
    ]
    x.merge(y)
    assert x.data == z


@pytest.mark.parametrize(
    ("merge_length", "success"), [(2, False), (4, False), (1, True), (3, True)]
)
def test_sequence_merge_raises_on_different_lengths(merge_length, success):
    x = NestedData(
        [
            {"color": "red", "value": "#f00"},
            {"color": "green", "value": "#0f0"},
            {"color": "blue", "value": "#00f"},
        ]
    )
    y = NestedData([{"realm": "earth"} for _ in range(merge_length)])
    z = [
        {"color": "red", "value": "#f00", "realm": "earth"},
        {"color": "green", "value": "#0f0", "realm": "earth"},
        {"color": "blue", "value": "#00f", "realm": "earth"},
    ]

    if not success:
        with pytest.raises(ValueError):
            x.merge(y)
    else:
        x.merge(y)
        assert x.data == z


def test_deep_sequence_can_be_merged_with_path():
    x = NestedData(
        {
            "colors": [
                {"color": "red", "value": "#f00"},
                {"color": "green", "value": "#0f0"},
                {"color": "blue", "value": "#00f"},
            ]
        }
    )
    y = NestedData([{"realm": "earth"} for _ in range(len(x))])
    z = [
        {"color": "red", "value": "#f00", "realm": "earth"},
        {"color": "green", "value": "#0f0", "realm": "earth"},
        {"color": "blue", "value": "#00f", "realm": "earth"},
    ]
    x.merge(y, path="colors")
    assert x.data["colors"] == z


def test_deep_sequence_can_be_merged():
    x = NestedData(
        {
            "colors": [
                {"color": "red", "value": "#f00"},
                {"color": "green", "value": "#0f0"},
                {"color": "blue", "value": "#00f"},
            ]
        }
    )
    y = NestedData({"colors": [{"realm": "earth"} for _ in range(len(x["colors"]))]})
    z = [
        {"color": "red", "value": "#f00", "realm": "earth"},
        {"color": "green", "value": "#0f0", "realm": "earth"},
        {"color": "blue", "value": "#00f", "realm": "earth"},
    ]
    x.merge(y, path="colors")
    assert x.data["colors"] == z


def test_sequence_accepts_slices_as_indexes_for_reading():
    x = NestedData([{"b": i} for i in range(5)])
    assert x[2:4].data == x.data[2:4]


def test_sequence_accepts_slices_as_indexes_for_deleting():
    x = NestedData([{"b": i} for i in range(5)])
    y = deepcopy(x.data)
    del y[2:4]
    del x[2:4]
    assert x.data == y


def test_sequence_accepts_slices_as_indexes_for_writting():
    x = NestedData([{"b": i} for i in range(5)])
    y = [{"b": 10}, {"b": 11}]
    z = deepcopy(x.data)
    x[2:4] = deepcopy(y)
    z[2:4] = y
    assert x.data == z


def test_nested_data_can_handle_path_components_as_tuples_1():
    a = NestedData({("person.address.city"): "São Paulo"})
    assert a["person", "address", "city"] == "São Paulo"


def test_nested_data_can_handle_path_components_as_tuples_2():
    a = NestedData({("person", "address", "city"): "São Paulo"})
    assert a["person.address.city"] == "São Paulo"


def test_nested_data_can_handle_path_components_as_tuples_3():
    a = NestedData({("person", "address"): []})
    a["person", "address"].append("test")
    assert a["person", "address", "0"]
    assert a["person", "address", 0]



import threading
from time import sleep
from unittest.mock import patch

from extradict.binary_tree_dict import PlainNode, AVLNode, TreeDict

import pytest


def test_node_value_equals_key_if_not_given():
    n = PlainNode(0)
    assert n.value == 0
    assert not n.left
    assert not n.right
    assert n.key_func is None
    assert n.leaf


def test_node_insert_right():
    n = PlainNode(0)
    n.insert(10)
    assert n.right.value == 10
    assert not n.left
    assert not n.leaf


def test_node_insert_left():
    n = PlainNode(0)
    n.insert(-10)
    assert n.left.value == -10
    assert not n.right
    assert not n.leaf


def test_node_insert_right_right():
    n = PlainNode(0)
    n.insert(10)
    n.insert(20)
    assert n.right.right.value == 20


def test_node_insert_right_left():
    n = PlainNode(0)
    n.insert(10)
    n.insert(5)
    assert n.right.left.value == 5


def test_node_insert_both_sides():
    n = PlainNode(0)
    n.insert(10)
    n.insert(-10)
    n.insert(5)
    assert n.right.value == 10
    assert n.left.value == -10
    assert n.right.left.value == 5


@pytest.mark.parametrize(
    ("values", "expected"),
    [
        ((0,), 1),
        ((0, 5), 2),
        ((0, 5, -5), 2),
        ((0, 5, -5, 7), 3),
        ((0, 5, -5, 7, 10), 4),  # not auto balancing
        ((0, 5, -5, 7, 10, -7), 4),
    ],
)
def test_depth(values, expected):
    root = PlainNode(values[0])
    for v in values[1:]:
        root.insert(v)
    assert root.depth == expected


def test_node_len():
    n = PlainNode(0)
    assert len(n) == 1
    n.insert(10)
    assert len(n) == 2
    n.insert(5)
    assert len(n) == 3
    n.insert(-10)
    assert len(n) == 4


def test_node_delete_leaf():
    n = PlainNode(0)
    n.insert(5)
    n.delete(5)
    assert not n.right
    assert not n.left


def test_node_delete_root_single_node():
    n = PlainNode(0)
    assert not (n.delete(0))


def test_node_delete_root_one_branch():
    n = PlainNode(0)
    n.insert(5)
    n.delete(0)
    assert n.value == 5
    assert n.depth == 1
    assert not n.left and not n.right


def test_node_delete_root_one_branch_to_left():
    n = PlainNode(0)
    n.insert(-5)
    n.delete(0)
    assert n.value == -5
    assert n.depth == 1
    assert not n.left and not n.right


def test_node_delete_root_two_branches():
    n = PlainNode(0)
    n.insert(5)
    n.insert(-5)
    n.delete(0)
    assert n.value in (5, -5)
    assert bool(n.left) ^ bool(n.right)


def test_node_delete_branch():
    n = PlainNode(0)
    n.insert(5)
    n.insert(10)
    n.delete(5)
    assert n.value is 0
    assert n.right.value == 10
    assert n.right.depth == 1
    assert not n.right.left and not n.right.right


def test_node_delete_branch_left():
    n = PlainNode(0)
    n.insert(5)
    n.insert(3)
    n.delete(5)
    assert n.value is 0
    assert n.right.value == 3
    assert n.right.depth == 1
    assert not n.right.left and not n.right.right


def test_node_delete_branch_2_leaves():
    n = PlainNode(0)
    n.insert(5)
    n.insert(10)
    n.insert(3)
    n.delete(5)
    assert n.value is 0
    assert n.right.value in (10, 3)
    assert n.right.depth == 2
    assert (
        n.right.left
        and n.right.left.value == 3
        or n.right.right
        and n.right.right.value == 10
    )
    assert bool(n.right.left) ^ bool(n.right.right)


def test_node_delete_non_existing_value():
    n = PlainNode(0)
    n.insert(5)
    n.insert(10)
    n.insert(3)
    n.insert(-5)
    with pytest.raises(KeyError):
        n.delete(7)


def test_node_get_root_identity():
    n = PlainNode(0)
    assert n.get(0) is n


def test_node_get_root_value():
    n = PlainNode(0, "zero")
    assert n.get(0).value == "zero"


def test_node_get_leave_identity():
    n = PlainNode(0)
    n.insert(5)
    assert n.get(5) is n.right


def test_node_get_leave_value():
    n = PlainNode(0, "zero")
    n.insert(5, "five")
    assert n.get(5).value == "five"


def test_node_get_branch_identity():
    n = PlainNode(0)
    n.insert(5)
    n.insert(10)
    assert n.get(5) is n.right


def test_node_get_branch_value():
    n = PlainNode(0, "zero")
    n.insert(5, "five")
    n.insert(10, "ten")
    assert n.get(5).value == "five"


def test_node_get_non_existing_element_raises_keyerror():
    n = PlainNode(0)
    n.insert(5)
    n.insert(10)
    n.insert(7)
    with pytest.raises(KeyError) as error:
        n.get(8)
    assert True


@pytest.mark.parametrize(
    ["nodes_to_insert", "search_value", "expected_return_values"],
    [
        ((0,), 0, (0, 0)),
        ((0,), 1, (0, None)),
        ((0,), -2, (None, 0)),
        ((0, 5), 2, (0, 5)),
        ((0, 5, 3), 2, (0, 3)),
        ((0, 5, 3), 3, (3, 3)),
        ((0, 5, 3), 3.5, (3, 5)),
        ((0, 5, 10, 7), 6, (5, 7)),
        ((0, 5, 10, 7), 8, (7, 10)),
        ((0, 5, 10, 7), 7, (7, 7)),
        ((0, 5, 15, 20, 10, 12, 13), 14, (13, 15)),
        ((0, -5, -3), 2, (0, None)),
        ((0, -5, -3, -4, -4.5), -4.2, (-4.5, -4)),
    ],
)
def test_node_get_non_existing_element_returns_path_to_closest(
    nodes_to_insert, search_value, expected_return_values
):
    values = iter(nodes_to_insert)
    n = PlainNode(next(values))
    for value in values:
        n.insert(value)

    closest_nodes = n.get_closest(search_value)

    assert closest_nodes[0].value == expected_return_values[0]
    assert closest_nodes[1].value == expected_return_values[1]


def test_node_get_identity_or_value_independs_other_nodes():
    n = PlainNode(0, "zero")
    n.insert(5)
    n.insert(10, "ten")
    assert n.get(5).value == 5


def test_node_insert_replaces_value():
    n = PlainNode(0, "zero")
    n.insert(5, "five")
    n.insert(10, "ten")
    assert n.get(5) is n.right
    assert n.get(5).value == "five"
    assert len(n) == 3
    n.insert(5, "fake")
    assert n.get(5) is n.right
    assert n.get(5).value == "fake"
    assert len(n) == 3


def test_node_insert_can_create_new():
    n = PlainNode(0, "zero")
    n.insert(5, "five")
    n.insert(10, "ten")
    assert n.get(5).value == "five"
    assert len(n) == 3
    n.insert(5, "fake", replace=False)
    assert n.get(5).value == "five"  # Retrieves first inserted value
    assert len(n) == 4
    assert n.get(5).right.left.value == "fake"


def test_node_iter():
    n = PlainNode(0)
    n.insert(5)
    n.insert(-5)
    n.insert(-3)
    assert [x.value for x in n] == [-5, -3, 0, 5]


@pytest.mark.parametrize(
    ["nodes_to_insert", "expected"],
    [
        [(0,), True],
        [(0, 1), True],
        [(0, 1, 2), False],
        [(0, 1, 2, 3), False],
        [(0, 10, 20, 25), False],
        [(0, 10, 20, 25, -10), False],
        [(0, 10, 20, 25, -10, -20), True],
        [(0, 10, 20, 25, -10, -20, -30), True],
        [(0, 10, 20, 25, -10, -20, -30, -40), True],
        [(0, 10, 20, 25, -10, -20, -30, -40, -50, -60), False],
    ],
)
def test_node_reports_balanced(nodes_to_insert, expected):
    values = iter(nodes_to_insert)
    n = PlainNode(next(values))
    for value in values:
        n.insert(value)

    assert n.balanced == expected


@pytest.mark.parametrize(
    ["nodes_to_insert", "expected_root"],
    [
        [(0,), 0],
        [
            (
                0,
                1,
                2,
            ),
            1,
        ],
        [(0, 1, 2, 3), 1],
        [(0, 10, 20, 25), 10],
        [(0, -10, -20, -30), -10],
        [(0, -10, -20, -15, -30), -10],
        [(0, 10, 20, -10, 25), 10],
        [(0, 10, 20, -10, -20, -30, -40, 25), -10],
        [(0, 10, 20, -10, -20, -30, -40, -50, -60, -70, -80, -90, 25), -50],
    ],
)
def test_avlnode_always_balanced(nodes_to_insert, expected_root):
    values = iter(nodes_to_insert)
    n = AVLNode(next(values))
    for value in values:
        n.insert(value)

    assert [node.value for node in n] == sorted(nodes_to_insert)
    assert n.value == expected_root
    assert n.balanced


@pytest.mark.parametrize(
    ["nodes_to_insert", "key", "expected"],
    [
        [(0, 4, 2, 1, 3), 3, (0, 4, 2, 3)],
        [(0, 4, 2, 1, 3), 0, (0,)],
        [(0, 4, 2, 1, 3), 4, (0, 4)],
        [(0, 4, 2, 1, 3), 3.5, (0, 4, 2, 3, None)],
        [(0, 4, 2, 1, 3), 0.5, (0, 4, 2, 1, None)],
        [(0, -4, -2), -2, (0, -4, -2)],
        [(0, -4, -2), -2.5, (0, -4, -2, None)],
    ],
)
def test_get_node_path_works(nodes_to_insert, key, expected):
    values = iter(nodes_to_insert)
    n = PlainNode(next(values))
    for value in values:
        n.insert(value)

    result = [node.value for node in n.get_node_path(key)]
    assert result == list(expected)


@pytest.mark.parametrize(
    ["nodes_to_insert", "node_to_start", "side", "expected_last_node"],
    [
        [(0, 5, 1, 2), 0, "right", 1],
        [(0, 5, 1, 2), 1, "right", 2],
        [(0, 5, 1, 2), 2, "right", 5],
        [(0, 5, 1, 2), 5, "right", None],
        [(0, -5, -1, -2), -1, "left", -2],
        [(0, -5, -1, -2), -2, "left", -5],
        [(0, -5, -1, -2), -5, "left", None],
        [(0, -5, -1, -2, 5, 1, 2), 2, "left", 1],
        [(0, -5, -1, -2, 5, 1, 2), 1, "left", 0],
        [(0, -5, -1, -2, 5, 1, 2), 0, "left", -1],
        [(0, -5, -1, -2, 5, 1, 2), -1, "left", -2],
    ],
)
def test_traverse_to_side(nodes_to_insert, node_to_start, side, expected_last_node):
    values = iter(nodes_to_insert)
    n = PlainNode(next(values))
    for value in values:
        n.insert(value)

    path = n._traverse_to_side(n.get_node_path(node_to_start), side)
    if not path:
        assert expected_last_node is None
        return
    assert path[-1].value == expected_last_node


@pytest.mark.parametrize(
    ["nodes_to_insert", "slice_", "expected"],
    [
        [(0, 1, 2, 3, 4, 5), (0, 2, None), (0, 1)],
        [(0, 1, 2, 3, 4, 5), (0, 1, None), (0,)],
        [(0, 1, 2, 3, 4, 5), (0, 3, None), (0, 1, 2)],
        [(0, 1, 2, 3, 4, 5), (0, 3, 1), (0, 1, 2)],
        [(0, 1, 2, 3, 4, 5), (3, 0, -1), (3, 2, 1)],
        [(0, 1, -1, 2, -2, 3, -3, 4, -4, 5, -5), (0, 3, 1), (0, 1, 2)],
        [(0, 1, -1, 2, -2, 3, -3, 4, -4, 5, -5), (-3, 0, 1), (-3, -2, -1)],
        [(0, 1, 2, 3, 4, 5), (-2, 2, None), (0, 1)],
        [(0, 1, 2, 3, 4, 5), (4, 7, None), (4, 5)],
        [(0, 1, 2, 3, 4, 5), (0.5, 3.5, None), (1, 2, 3)],
        [(0, 1, 2), (None, None, None), (0, 1, 2)],
        [(0, 1, -1), (None, None, None), (-1, 0, 1)],
        [(0, 1, -1), (None, None, -1), (1, 0, -1)],
        [(0, 1, 2, 3, 4, 5), (0, 2, -1), ()],
        [(0, 1, 2, 3, 4, 5), (2, 0, 1), ()],
    ],
)
def test_node_slice_works(nodes_to_insert, slice_, expected):
    values = iter(nodes_to_insert)
    n = AVLNode(next(values))
    for value in values:
        n.insert(value)

    result = [node.value for node in n.iter_slice(slice(*slice_))]
    assert result == list(expected)


@pytest.mark.parametrize(
    ["nodes_to_insert", "slice_", "expected"],
    [
        [(0, 1, 2, 3, 4, 5), (0, 4, 2), (0, 2)],
        [(0, 1, 2, 3, 4, 5, 6, 7, 8), (0, None, 3), (0, 3, 6)],
        [(0, 1, 2, 3, 4, 5, 6, 7, 8), (None, 1, -2), (8, 6, 4, 2)],
        [(0, 1, 2, 3, 4, 5, 6, 7, 8), (None, -1, -2), (8, 6, 4, 2, 0)],
        [(0, 1, 2, 3, 4, 5), (0.5, 5, 2), (1, 3)],
        [(0, 1, 2, 3, 4, 5), (0.5, 5.5, 2), (1, 3, 5)],
        [(0, 1, 2, 3, 4, 5, 6, 7, 8), (6.5, None, -2), (6, 4, 2, 0)],
    ],
)
def test_node_slice_works_with_larger_than_unity_step(
    nodes_to_insert, slice_, expected
):
    values = iter(nodes_to_insert)
    n = AVLNode(next(values))
    for value in values:
        n.insert(value)

    result = [node.value for node in n.iter_slice(slice(*slice_))]
    assert result == list(expected)


def delay_wrapper_decorator(func, delay=0.025):
    def wrapper(*args, **kw):
        sleep(delay)
        return func(*args, **kw)

    return wrapper


@pytest.mark.skip("test code incomplete")
def test_plainnode_avoids_thread_race_condition():
    pass


def test_treedict_key_function_works():
    words = "art bee cat desk elephant".split()
    rwords = ["".join(reversed(w)) for w in words]
    a = TreeDict(key=lambda k: list(reversed(k)))
    for word in rwords:
        a[word] = word

    assert list(a) == rwords


def test_treedict_clear_and_reinsert():
    a = TreeDict()
    a["a"] = "a"
    a["b"] = "b"
    a.clear()
    a["c"] = "c"
    a["d"] = "d"

    assert len(a) == 2
    assert "c" in a
    assert "a" not in a



# flake8: noqa

import enum
import threading

from extradict import MapGetter


def test_mapgetter_creates_local_variables():
    a = dict(b=1, c=2)
    with MapGetter(a) as a:
        from a import b, c
    assert b == 1 and c == 2


def test_mapgetter_can_be_used_with_key_renaming():
    a = dict(b=1, c=2)
    with MapGetter(a) as a:
        from a import b as d, c as e
    assert d == 1 and e == 2


def test_mapgetter_can_use_any_name():
    a = dict(b=1, c=2)
    with MapGetter(a) as anyname:  # flake8: noqa
        from anyname import b, c
    assert b == 1 and c == 2


def test_mapgetter_can_use_existing_module_name():
    a = dict(b=1, c=2)
    with MapGetter(a) as math:  # NOQA
        from math import b, c
    assert b == 1 and c == 2


def test_regular_import_works_from_within_():
    a = dict(b=1, c=2)
    with MapGetter(a) as a:  # NOQA
        from math import cos
    assert cos


def test_mapgetter_is_thread_safe():
    import time

    a = dict(b=1, c=2)

    nonlocal_check = [False, None, None, None]

    def thread_1():
        with MapGetter(a) as math:  # NOQA
            from math import b, c

            nonlocal_check[0] = True
            time.sleep(0.01)
        nonlocal_check[2] = b
        nonlocal_check[3] = c

    def thread_2():
        while not nonlocal_check[0]:
            time.sleep(0.001)
        from math import cos

        assert cos
        nonlocal_check[0] = False
        nonlocal_check[1] = cos

    t1 = threading.Thread(target=thread_1)
    t2 = threading.Thread(target=thread_2)
    t1.start()
    t2.start()
    t1.join()
    t2.join()
    from math import cos

    assert nonlocal_check == [False, cos, 1, 2]


def test_mapgetter_works_with_default_value():
    with MapGetter(default=None) as m:  # NOQA
        from m import n, o, p
    assert n is None
    assert o is None
    assert p is None


def test_mapgetter_works_with_default_function_with_parameters():
    with MapGetter(default=lambda name: name) as m:  # NOQA
        from m import foo, bar
    assert foo == "foo"
    assert bar == "bar"


def test_mapgetter_works_with_default_function_without_parameters():
    with MapGetter(default=lambda: "baz") as m:  # NOQA
        from m import foo, bar
    assert foo == "baz"
    assert bar == "baz"


def test_mapgetter_works_with_default_non_function_callable():
    class Defaulter(object):
        def __call__(self, name):
            return name

    with MapGetter(default=Defaulter()) as m:  # NOQA
        from m import foo, bar
    assert foo == "foo"
    assert bar == "bar"

    class Defaulter(object):
        def __call__(self):
            return "baz"

    with MapGetter(default=Defaulter()) as m:
        from m import foo, bar

    assert foo == "baz"
    assert bar == "baz"


def test_mapgetter_works_with_defaultdict():
    from collections import defaultdict

    with MapGetter(defaultdict(lambda: "baz")) as m:  # NOQA
        from m import foo, bar

    assert foo == "baz"
    assert bar == "baz"


def test_mapgetter_works_with_mapping_and_default_parameter():
    a = dict(b=1, c=2)
    with MapGetter(a, default=lambda name: name) as a: # NOQA
        from a import b, c, d
    assert b == 1 and c == 2 and d == "d"


def test_mapgetter_accepts_import_object_attributes():
    try:
        from types import SimpleNameSpace
    except ImportError:

        class SimpleNameSpace(object):
            pass

    test_obj = SimpleNameSpace()
    test_obj.a = 1
    test_obj.b = 2
    with MapGetter(test_obj):
        from test_obj import a, b

    assert a == test_obj.a
    assert b == test_obj.b


def test_mapgetter_works_with_enums():
    class A(enum.Enum):
        foo = 0
        bar = 1
        baz = 2

    with MapGetter(A) as A:
        from A import foo, bar, baz

    assert foo is A.foo
    assert bar is A.bar
    assert baz is A.baz
    assert foo.value == 0



import pytest

from extradict import BijectiveDict


def test_bijective_works():
    a = BijectiveDict(a=1)
    assert a["a"] == 1
    assert a[1] == "a"
    assert len(a) == 2
    assert set(iter(a)) == {"a", 1}


def test_bijective_dict_raises_on_unhashable_value():
    with pytest.raises(TypeError):
        BijectiveDict["a"] = []


def test_bijective_changes_reciprocal_value_on_assignement():
    a = BijectiveDict(a=1)
    a[1] = "b"
    assert a["b"] == 1
    assert "a" not in a


def test_bijective_keeps_assigned_keys():
    a = BijectiveDict(a=1)
    assert set(a.assigned_keys) == {"a"}
    assert set(a.generated_keys) == {1}


def test_bijective_sub_dicts_works():
    a = BijectiveDict(a=1)
    assert a.assigned == {"a": 1}
    assert a.generated == {1: "a"}


def test_bijective_tracks_assigned_keys():
    a = BijectiveDict(a=1)
    a[1] = "b"
    assert set(a.assigned_keys) == {1}
    assert set(a.generated_keys) == {"b"}
    assert "a" not in a.assigned_keys


def test_bijective_repr():
    a = BijectiveDict(a=1)
    assert (
        repr(a) == "BijectiveDict({'a': 1, 1: 'a'})"
        or "BijectiveDict({1: 'a', 'a': 1})"
    )


def test_bijective_copy():
    a = BijectiveDict(a=1)
    b = a.copy()
    assert b == a
    assert b is not a



import sys
import threading

_sentinel = object()


class _PseudoModule(object):
    def __init__(self, mapping, default=_sentinel):
        try:
            self.__dict__ = mapping
        except TypeError:
            self._obj = mapping
        else:
            self._obj = _sentinel
        self._default = default

    def getlines(self, file):
        return ""

    def __getattr__(self, attr):
        if self._obj is not _sentinel:
            try:
                return getattr(self._obj, attr)
            except AttributeError:
                if self._default is _sentinel:
                    raise AttributeError(
                        f"Object '{type(self._obj)}' has no attribute '{attr}' "
                    )
        if self._default is _sentinel:
            # Try to extract a value from mapping, which could be a
            # defaultdict or some similar mapping:
            try:
                value = self.__dict__[attr]
            except KeyError:
                raise AttributeError(
                    f"Mapping '{type(self.mapping)}' has no '{attr}' key"
                )
            return value
        if not callable(self._default):
            return self._default
        try:
            value = self._default(attr)
        except TypeError:
            # try default dict style argumentless default factory:
            value = self._default()
        return value


class MapGetter(object):
    """
    A context manager to allow one to "import" variables from a mapping or
    factory function. This helps preserve DRY principle:

    # Example:
    >>> a = dict(b=1, c=2)
    >>> with MapGetter(a) as blah:
    ...    from blah import b, c

    >>> print((b, c))
    (1, 2)

    It is intesresting to note that it will work for ordinary attributes
    from Python objects, and, as well, for constant names inside a Python
    enum.Enum class. That may be the major use case of this:

    In[1]: import enum
    In [2]: from extradict import MapGetter

    In 43]: class Colors(enum.Enum):
        ...:     RED = 1, 0, 0
        ...:     GREEN = 0, 1, 0
        ...:     BLUE = 0, 0, 1
        ...:

    In [4]: with MapGetter(Colors):
        ...:     from Colors import RED, GREEN, BLUE
        ...:

    In [5]: RED, GREEN, BLUE
    Out[5]: (<Colors.RED: (1, 0, 0)>, <Colors.GREEN: (0, 1, 0)>, <Colors.BLUE: (0, 0, 1)>)


    """

    def __init__(self, mapping=None, default=_sentinel):
        if mapping is None and default is _sentinel:
            raise TypeError(
                "MapGetter must be called with at least one of mapping or default (value/factory function)"
            )
        self.builtins = (
            __builtins__ if isinstance(__builtins__, dict) else __builtins__.__dict__
        )
        self.mapping = mapping if mapping is not None else {}
        self.default = default

    def __enter__(self):
        self.original_import = self.builtins["__import__"]
        self.builtins["__import__"] = self._map_getter
        self._thread = threading.current_thread()
        return self.mapping

    def _map_getter(self, name, globals_, locals_, from_list, level=-1):
        # "from_list" gets the names wanted in the "import ... from ... " statement,
        # however, they are not actually expected to be returned.
        # the fact the remainer of the import machinery does not use them
        # makes it worth to keep this implementation using a __import__ monkey patch -
        # no other func gets "from_list" and all the machinery expects the actual
        # module to be created on sys.modules - which we don't want.

        # FIXME: Make this asyncio safe as well.
        # (May wait for Python 3.7 with PEP-555 implemented)
        if (
            threading.current_thread() != self._thread
            or sys._getframe().f_back.f_locals.get(name, None) is not self.mapping
        ):
            return self.original_import(name, globals_, locals_, from_list, level)
        return _PseudoModule(self.mapping, self.default)

    def __exit__(self, type, value, traceback):
        self.builtins["__import__"] = self.original_import
        return False



try:
    from collections.abc import MutableMapping
except ImportError:
    from collections import MutableMapping

sentinel = object()


class Behaviors:
    """Enumeration to dictate the possible behaviors of a BijectiveDict."""

    for name in ("single_permissive", "single_strict", "multi"):
        locals()[name] = name


class BijectiveDict(MutableMapping):
    """Mapping that auto-reflects all key->value pairs as value->key.

    This is a bijective dictionary for which each pair key, value added
    is also added as value, key.

    The explictly inserted keys can be retrieved as the "assigned_keys"
    attribute - and a dictionary copy with all such keys is available
    at the "BijectiveDict.assigned".
    Conversely, the generated keys are exposed as "BijectiveDict.generated_keys"
    and can be seen as a dict at "Bijective.generated
    """

    def __init__(self, *args, **kw):
        data = dict(*args, **kw)
        self._data = dict()
        self.assigned_keys = set()
        for key, value in data.items():
            self[key] = value

    def copy(self, version=None):
        return self.__class__(self.assigned)

    def __getitem__(self, item):
        return self._data[item]

    def __setitem__(self, key, value):
        reciprocal = self._data.get(key, sentinel)
        if reciprocal is not sentinel:
            del self[key]
        self.assigned_keys.add(key)
        self._data[key] = value
        self._data[value] = key

    def __delitem__(self, key):
        reciprocal = self[key]
        try:
            self.assigned_keys.remove(key)
        except KeyError:
            pass
        try:
            self.assigned_keys.remove(reciprocal)
        except KeyError:
            pass
        del self._data[key]
        del self._data[reciprocal]

    def __iter__(self):
        return iter(self._data)

    def __len__(self):
        return len(self._data)

    @property
    def generated_keys(self):
        return set(self._data.keys()) - self.assigned_keys

    @property
    def assigned(self):
        return {key: self[key] for key in self.assigned_keys}

    @property
    def generated(self):
        return {key: self[key] for key in self.generated_keys}

    def __repr__(self):
        return "{}({})".format(self.__class__.__name__, repr(self._data))


class ReversibleDict(MutableMapping):
    """Mapping that tracks all key->value assignments in a .generated attribute


    This is a bijective dictionary for which each pair key, value added
    is also added available as  value, key.

    The explictly inserted keys can be retrieved as the "assigned_keys"
    attribute - and a dictionary copy with all such keys is available
    at the "BijectiveDict.assigned".
    Conversely, the generated keys are exposed as "BijectiveDict.generated_keys"
    and can be seen as a dict at "Bijective.generated
    """

    def __init__(self, *args, **kw):
        data = dict(*args, **kw)
        self.assigned = dict()
        self.generated = dict()
        for key, value in data.items():
            self[key] = value

    def copy(self, version=None):
        return self.__class__(self.assigned)

    def __getitem__(self, item):
        return self._data[item]

    def __setitem__(self, key, value):
        reciprocal = self._data.get(key, sentinel)
        if reciprocal is not sentinel:
            del self[key]
        self.assigned_keys.add(key)
        self._data[key] = value
        self._data[value] = key

    def __delitem__(self, key):
        reciprocal = self[key]
        try:
            self.assigned_keys.remove(key)
        except KeyError:
            pass
        try:
            self.assigned_keys.remove(reciprocal)
        except KeyError:
            pass
        del self._data[key]
        del self._data[reciprocal]

    def __iter__(self):
        return iter(self._data)

    def __len__(self):
        return len(self._data)

    @property
    def generated_keys(self):
        return set(self._data.keys()) - self.assigned_keys

    @property
    def assigned(self):
        return {key: self[key] for key in self.assigned_keys}

    @property
    def generated(self):
        return {key: self[key] for key in self.generated_keys}

    def __repr__(self):
        return "{}({})".format(self.__class__.__name__, repr(self._data))



# coding: utf-8
"""
Module for a versioned version of Python's dictionaries -
one can retrieve any "past" value that had existed previously
by using the "get" method with an extra "version" parameter -
also, the "version" is an explicit read only attribute
that allows one to know wether a dictionary had  changed.

"""

import threading
from collections import namedtuple
from copy import copy

try:
    from collections.abc import MutableMapping
except ImportError:
    from collections import MutableMapping


VersionedValue = namedtuple("VersionedValue", "version value")

_Deleted = object()


class VersionDict(MutableMapping):
    _dictclass = dict

    def __init__(self, *args, **kw):
        self._version = 0
        initial = self._dictclass(*args, **kw)
        self.data = self._dictclass()
        for key, value in initial.items():
            self.data[key] = [VersionedValue(self._version, value)]
        self.local = threading.local()
        self.local._updating = False
        self._init_lock()

    def _init_lock(self):
        self.lock = threading.RLock()

    def copy(self, version=None):
        new = VersionDict.__new__(self.__class__)
        new._init_lock()
        if version is None or version >= self.version:
            new._version = self._version
            new.data = copy(self.data)
        else:
            new._version = version
            new.data = self._dictclass()
            for key, value in self.data.items():
                new_values = [
                    value_item for value_item in value if value_item.version <= version
                ]
                if new_values:
                    new.data[key] = new_values

        return new

    def freeze(self, version=None):
        """Create a shallow copy of an specific version
        of the dictionary. If version is not given, creates
        a shallow copy of the current version.
        """
        new = self._dictclass()
        if version is None:
            version = self.version
        _default = object()
        for key, value in self.data.items():
            frozen_value = self.get(key, _default, version=version)
            if frozen_value is not _default:
                new[key] = frozen_value
        return new

    def update(self, other):
        """The update operation uses a single version number for
        all affected keys
        """
        with self.lock:
            self._version += 1
            try:
                self.local._updating = True
                super(VersionDict, self).update(other)
            finally:
                self.local._updating = False

    def get(self, item, default=_Deleted, version=None):
        """
        VersionedDict.get(item, default=None) -> same as dict.get
        VersionedDict.get(item, [default=Sentinel], version) ->
            returns existing value at the given dictionary version. If
            value was not set, and no default is given,
            raises KeyError (unlike regular dict)
        """
        if version is None:
            return super(VersionDict, self).get(
                item, default=(None if default is _Deleted else default)
            )
        try:
            values = self.data[item]
            i = -1
            while values[i].version > version:
                i -= 1
        except (KeyError, IndexError):
            if default is not _Deleted:
                return default
            raise KeyError("'{}' was not set at dict version {}".format(item, version))
        if values[i].value is _Deleted:
            if default is not _Deleted:
                return default
            raise KeyError("'{}' was not set at dict version {}".format(item, version))
        return values[i].value

    def __getitem__(self, item):
        value = self.data[item][-1]
        if value.value is _Deleted:
            raise KeyError("'{}' is a deleted key".format(item))
        return value.value

    def __setitem__(self, item, value):
        with self.lock:
            if not self.local._updating:
                self._version += 1
            if item not in self.data:
                self.data[item] = []
            self.data[item].append(VersionedValue(self._version, value))

    def __delitem__(self, item):
        with self.lock:
            self._version += 1
            self.data[item].append(VersionedValue(self._version, _Deleted))

    def __iter__(self):
        for key, value in self.data.items():
            if value[-1].value is not _Deleted:
                yield key

    def __len__(self):
        return sum(1 for x in self.data.values() if x[-1].value != _Deleted)

    @property
    def version(self):
        return self._version

    def __repr__(self):
        return "<{}({}) at version {}>".format(
            self.__class__.__name__,
            ", ".join("{}={!r}".format(*item) for item in self.items()),
            self.version,
        )


class OrderedVersionDict(VersionDict):
    _dictclass = dict

    def freeze(self, version=None):
        new = self._dictclass()
        if version is None:
            version = self.version
        # Making a versioned copy instead of iterating on self.data preserves
        # order semantics for OrderedVersionDict
        for key, value in self.copy(version).items():
            new[key] = value
        return new

    def __iter__(self):
        versions_for_keys = {}
        for key, values in self.data.items():
            if values[-1].value is _Deleted:
                continue
            versions_for_keys.setdefault(values[-1].version, []).append(key)
        for version in sorted(versions_for_keys.keys()):
            # yield from versions_for_keys[key]
            for key in versions_for_keys[version]:
                yield key



from extradict import VersionDict as VD

import pytest


@pytest.fixture
def vd():
    return VD(a=0, b=1, c=2)


def test_vd_retrieves_values(vd):
    assert vd["a"] == 0
    assert vd["b"] == 1
    assert vd["c"] == 2


def test_vd_new_dictionary_is_at_version_0(vd):
    assert vd.version == 0


def test_vd_increase_version_on_attr_change(vd):
    vd["a"] = 1
    assert vd["a"] == 1
    assert vd.version == 1


def test_vd_can_retrieve_previous_version_value(vd):
    vd["a"] = 1
    assert vd.get("a") == 1
    assert vd.get("a", version=0) == 0
    assert vd.get("a", version=1) == 1


def test_vd_can_retrieve_previous_version_value(vd):
    vd["a"] = 1
    assert vd.get("a") == 1
    assert vd.get("a", version=0) == 0
    assert vd.get("a", version=1) == 1


def test_vd_raises_keyerror(vd):
    with pytest.raises(KeyError):
        vd["d"]


def test_vd_update_increases_version_in_one_number(vd):
    pass



from __future__ import annotations

from collections import deque
from collections.abc import Mapping
import typing as T


SENTINEL = object()


class Queue:
    """Used as bucket class by Grouper. Works as an iterator value for each key in the Grouper mapping.

    Normally each element is consumed as it is fetched - but the elements
    are kept in a `collections.deque` instance in the .data attribute.
    One can access .data directly if desired.
    """

    __slots__ = ("data", "key", "parent")

    def __init__(self, parent: Grouper, key: T.Hashable):
        self.key = key
        self.parent = parent
        self.data = deque()

    def peek(self, default: T.Any = False) -> T.Any:
        """Return the next available value under this bucket, if any.
        If there is no waiting value, returns the passed in default value,
        or False, if none was given.

        Calling this won't consume the next element, neither cause
        the source iterator in the parent to advance.
        """
        return self.data[0] if self.data else default

    def __iter__(self):
        return self

    def __next__(self):
        if self.data:
            return self.data.popleft()
        try:
            self.parent.fetch_next(self.key)
        except StopIteration:
            raise
        else:
            return self.data.popleft()

    def append(self, value: T.Any) -> None:
        """Used internally by the parnt grouper to feed the filtered data"""
        self.data.append(value)

    def __repr__(self):
        return f"Queue <{self.data}>"


class Grouper(Mapping):
    """Grouper mapping:

    Think of it as an itertools.groupby which returns a mapping
    Or as an itertools.tee that splits the stream into filtered
    substreams according to the passed key-callable.

    Given an iterable and a key callable,
    each element in the iterable is run through the key callable and
    made available in an iterator under a bucket using the resulting key-value.

    The source iterable need not be ordered (unlike itertools.groupby).
    If no key function  is given, the identity function is used.

    The items will be made available under the iterable-values as requested,
    in a lazy way when possible. Note that several different method calls may
    precipatate an eager processing of all items in the source iterator:
    .keys() or len(), for example.

    Whenever a new key is found during input consumption, a "Queue" iterator,
    which is a thin wrapper over collections.deque is created under that key
    and can be further iterated to retrieve more elements that map to
    the same key.

    In short, this is very similar to `itertools.tee`, but with a filter
    so that each element goes to a mapped bucket.

    Once created, the resulting object may obtionally be called. Doing this
    will consume all data in the source iterator at once, and return a
    a plain dictionary with all data fetched into lists.

    For example, to divide a sequence of numbers from 0 to 10 in
    5 buckets, all one need to do is: `Grouper(myseq, lambda x: x // 2)`

    Or:
    even_odd = `Grouper(seq, lambda x: "even" if not x % 2 else "odd")`

    """

    def __init__(
        self,
        source: T.Union[T.Sequence, T.Iterable],
        key: T.Callable[[T.Any], T.Hashable] = None,
    ):
        self.key = key if key is not None else lambda x: x
        self.source = iter(source)
        self.data = dict()

    def __getitem__(self, key: T.Hashable):
        if key not in self.data:
            SENTINEL2 = object()
            if self.fetch_next(key, SENTINEL2) is SENTINEL2:
                raise KeyError(key)
        return self.data[key]

    def fetch_next(self, key: T.Hashable, default: T.Any = SENTINEL):
        """Advances consuming the source until a new value for 'key' is fetched.

        When source is exhausted, either raises StopIteration or returns the
        default value, if it is given.

        Used internally by the Queue iterators to advance to their next values
        """
        for new_key, value in self.advance():
            if key == new_key:
                return value
        if default is SENTINEL:
            raise StopIteration
        return default

    def advance(self):
        """A generator that consumes one item from source, feeds the internal buckets
        and yields the (key, value) pair each time it is 'nexted'.

        This is used internally as the mechanism to fill the bucket queues.
        If there the intent is just to consume the source and get the key, value pair
        without storing the values in the buckets, use ".consume()"
        """

        for key, value in self.consume():
            if key not in self.data:
                self.data[key] = Queue(self, key)
            self.data[key].append(value)
            yield key, value

    def consume(self):
        """A generator that consumes one item from source, feeds the internal buckets
        and yields the (key, value) pair each time it is 'nexted'.

        This is used internally as the mechanism to advance the source generator
        and get the corresponding key. The values are not stored for further use,
        and just yielded as key, value pair.

        This can be used directly, but one might just use
        `map(lambda v: (key(v), v), source)` instead of using a Grouper object.

        If this is called directly when trying to make use of the buckets undefined
        results will ensue.
        """
        for value in self.source:
            key = self.key(value)
            yield key, value

    def __iter__(self):
        self.consume_all()
        return iter(self.data)

    def __len__(self):
        self.consume_all()
        return len(self.data.keys())

    def consume_all(self):
        # Consumes all the remaining source
        # this "hidden" recipe is the most efficient
        # way in the cPython interpreter to consume a generator
        # - check: https://docs.python.org/3/library/itertools.html#itertools-recipes
        deque(self.advance(), maxlen=0)

    def __call__(self, keyhint: T.Container[str] = ()) -> dict[T.Hashable, list[T.any]]:
        """Consumes all the source iterator, and returns a plain dictionay with
        all elements inside lists under the appropriate keys. If keyhint
        is passed, keys for which there are no elements are created with
        empty lists. (But extra keys yielded by the key function and
        not present in keyhint will be present in the result, nonetheless).

        The iterators in the Grouper object itself are not themselves consumed
        (although the expected pattern if one calls this is that the resulting
        dictionary is used and the grouper object is discarded)

        """
        keyhint = set(keyhint)
        result = {key: list(self[key].data) for key in self}
        for remaining in keyhint - result.keys():
            result[remaining] = []
        return result

    def __repr__(self):
        return f"Grouper by {self.key}"



from collections.abc import Mapping, MutableMapping
from copy import copy
from itertools import zip_longest
from operator import le as le_op, ge as ge_op

"""Implements an AVLTree auto=balancing tree with a Python Mapping interface"""


_empty = object()


class EmptyNode:
    __slots__ = ()
    depth = 0
    value = None
    key = None

    def __bool__(self):
        return False

    def __len__(self):
        return 0

    def __repr__(self):
        return "EmptyNode"

    def __str__(self):
        return ""

    def __iter__(self):
        return iter(())

    def get(self, key):
        raise KeyError(key)

    def delete(self, key):
        raise KeyError(key)


EmptyNode = EmptyNode()  # The KISS Singleton


class PlainNode:
    __slots__ = "key value _left _right key_func _len _depth".split()

    def __init__(
        self, key, value=_empty, left=EmptyNode, right=EmptyNode, key_func=None
    ):
        self.key = key
        self.value = value if value != _empty else key
        self.left = left
        self.right = right
        self.key_func = key_func
        self._depth = 1
        self._len = 1

    @property
    def leaf(self):
        return not (self.left or self.right)

    @property
    def depth(self):
        return self._depth

    @property
    def balanced(self):
        return abs(self.left.depth - self.right.depth) <= 1

    @property
    def right(self):
        return getattr(self, "_right", EmptyNode)

    @right.setter
    def right(self, value):
        self._right = value
        self._update_depth()
        self._update_len()

    @property
    def left(self):
        return getattr(self, "_left", EmptyNode)

    @left.setter
    def left(self, value):
        self._left = value
        self._update_depth()
        self._update_len()

    def insert(self, key, value=_empty, replace=True):
        self_key = self._cmp_key(self.key)
        new_key = self._cmp_key(key)
        if new_key == self_key and replace:
            self.value = value if value is not _empty else key
            if self.key != key:
                self.key = key
            return
        try:
            target_str = "left" if new_key < self_key else "right"
        except TypeError:
            raise KeyError(
                f"{key!r} {'was converted to ' if type(key) != type(new_key) else 'has'}type {type(new_key)} and cant't be compared with {type(self_key)} "
            )
        target = getattr(self, target_str)
        if target:
            target.insert(key, value, replace)
        else:
            setattr(
                self,
                target_str,
                type(self)(key=key, value=value, key_func=self.key_func),
            )
        self._update_depth()
        self._update_len()

    def get(self, key):
        self_key = self._cmp_key(self.key)
        new_key = self._cmp_key(key)

        if self_key == new_key:
            return self
        try:
            new_key_smaller = new_key < self_key
        except TypeError:
            raise KeyError(f"{key} type incompatible with other keys in the tree")

        if new_key_smaller:
            return self.left.get(key)
        return self.right.get(key)

    def delete(self, key):
        self_key = self._cmp_key(self.key)
        new_key = self._cmp_key(key)

        if self_key == new_key:
            if self.leaf:
                return EmptyNode
            target_str = "left" if self.left.depth > self.right.depth else "right"
            target = getattr(self, target_str)
            self._mute_into(target)
            key = target.key  # key to delete in target subtree.
        else:
            target_str = "left" if new_key < self_key else "right"
            target = getattr(self, target_str)

        setattr(self, target_str, target.delete(key))
        self._update_depth()
        self._update_len()
        return self

    def __iter__(self):
        yield from self.left
        yield self
        yield from self.right

    def __len__(self):
        # This and .depth() where built as lazy properties,
        # iterating on childs in an earlier version - but
        # that can REALLY slow things down.
        # No "_len" or "_depth", no "_update_*" calls... but O(N²) instead of O(N).

        # Refactoring to an 'observer' pattern may yet take place.

        # return len(self.right) + len(self.left) + 1

        return self._len

    def iter_slice(self, slice_):
        step = slice_.step if slice_.step is not None else 1

        seek_direction = "right" if step > 0 else "left"
        start_side = "left" if step > 0 else "right"
        operator = ge_op if seek_direction == "right" else le_op

        start, start_fallback = (
            self.get_closest(slice_.start)
            if slice_.start is not None
            else (None, self._get_extreme_node(start_side))
        )
        start = start or start_fallback
        start_key = start._cmp_key(start.key)
        slice_start_key = start._cmp_key(slice_.start)

        if slice_.stop is not None and operator(start_key, start._cmp_key(slice_.stop)):
            return

        if slice_.start is None or operator(start_key, slice_start_key):
            yield start
            counter = 0
        else:
            counter = -1

        path = self.get_node_path(start.key)
        while True:
            path = self._traverse_to_side(path, side=seek_direction)
            node = path[-1] if path else EmptyNode
            if not node or (
                slice_.stop is not None
                and operator(node._cmp_key(node.key), node._cmp_key(slice_.stop))
            ):
                break
            counter += 1
            if not counter % abs(step):
                yield node

    def get_node_path(self, key, path=None):
        """Retrieve a list of nodes down to the node with the given Key.

        If the key does not exist, the last element is set to EmptyNode
        """
        if path is None:
            path = []
        path.append(self)
        self_key = self._cmp_key(self.key)
        new_key = self._cmp_key(key)
        if self_key == new_key:
            return path
        if new_key < self_key:
            if self.left:
                return self.left.get_node_path(key, path)
        elif self.right:
            return self.right.get_node_path(key, path)
        path.append(EmptyNode)
        return path

    def get_closest(self, key, path=None):
        """Retrieves two closest nodes on the tree to given key.

        Returns:
            A 2 Tuple with Two Node instances;

        If Key is found in the tree, its node is returned twice - otherwise
        the two nodes that are closest to it according to the comparation made
        by key_func. If the searched key would fall of the extreme right or
        extreme left of the Tree, the returned parent corresponding to "the abyss"
        is the EmptyNode singleton.
        """
        path = self.get_node_path(key, path)
        if path[-1]:  # key found - return twice the same node.
            return path[-1], path[-1]
        path.pop()
        closest = path[-1]
        closest_key = self._cmp_key(closest.key)
        new_key = self._cmp_key(key)

        if new_key > closest_key:
            return (
                closest,
                closest._get_closest_ancestor_on_other_side(path, side="right")[1],
            )
        return (
            closest._get_closest_ancestor_on_other_side(path, side="left")[1],
            closest,
        )

    def _cmp_key(self, key):
        return self.key_func(key) if self.key_func else key

    def _get_extreme_node(self, side):
        node = self
        while node:
            parent = node
            node = getattr(node, side)
        return parent

    def _get_closest_ancestor_on_other_side(self, path, side):
        # backtrack up to detour
        index = len(path) - 1
        while True:
            index -= 1
            if index < 0:
                return None, EmptyNode
            side_of_child = (
                "same" if getattr(path[index], side) is path[index + 1] else "other"
            )
            if side_of_child == "other":
                return index, path[index]

    def _update_depth(self):
        self._depth = max(self.left.depth, self.right.depth) + 1

    def _mute_into(self, other, full=False):
        self.key = other.key
        self.value = other.value
        self.key_func = other.key_func
        if full:
            self.right = other.right
            self.left = other.left

    def _update_len(self):
        self._len = 1 + len(self.left) + len(self.right)

    def _traverse_to_side(self, path, side):
        """Gets the next closest node key on the desired direction

        Search path is modified in place and returned with the new node
        as last element. Subsequent calls with the modified path
        will get the subsequent nodes in the same direction.

        """
        other_side = "left" if side == "right" else "right"
        # Descend on the searched_side:
        next_child = getattr(path[-1], side)
        found_descending = False
        while next_child:
            found_descending = True
            path.append(next_child)
            # now, try to keep on the "other_side" branch as far as possible.
            grand_child = getattr(next_child, other_side)
            while grand_child:
                path.append(grand_child)
                grand_child = getattr(grand_child, other_side)
            next_child = getattr(grand_child, side, None)
        if found_descending:
            return path

        current = path.pop()
        while path:
            if getattr(path[-1], other_side) is current:
                return path
            current = path.pop()
        return path

    def graph_repr(self, identifier="key"):
        str_repr = str(self.key) if identifier == "key" else f"{self.key}: {self.value}"
        repr_left = (self.left.graph_repr(identifier) if self.left else "").split("\n")
        repr_right = (self.right.graph_repr(identifier) if self.right else "").split(
            "\n"
        )
        len_l = len(repr_left[0])
        len_r = len(repr_right[0])
        width = len_l + len_r + 3
        result = [str_repr.center(width)]
        # result.append("/     \\".center(width))
        for line_left, line_right in zip_longest(repr_left, repr_right, fillvalue=""):
            result.append((line_left + "   " + line_right).center(width))
        return "\n".join(result)

    def __repr__(self):
        return f"{self.key}, ({repr(self.left) if self.left else ''}, {repr(self.right) if self.right else ''})"


class AVLNode(PlainNode):
    __slots__ = ()

    def insert(self, key, value=_empty, replace=True):
        super().insert(key, value, replace)
        self.balance()

    def delete(self, key):
        result = super().delete(key)
        self.balance()
        return result

    def balance(self):
        if self.balanced:
            return

        if self.left.depth > self.right.depth:
            self._avl_rotate_right()
        else:
            self._avl_rotate_left()

    def _avl_rotate_left(self):
        new_parent = self.right
        new_self = copy(self)
        new_self.right = new_parent.left
        new_parent.left = new_self
        self._mute_into(new_parent, full=True)

    def _avl_rotate_right(self):
        new_parent = self.left
        new_self = copy(self)
        new_self.left = new_parent.right
        new_parent.right = new_self
        self._mute_into(new_parent, full=True)


class TreeDict(MutableMapping):
    """Implements an AVLTree autobalancing tree with a Python Mapping interface.


    It can be used as a normal dictionary, although it will be much less eficient
    than a native dict.  Key types must not be mixed, or one will get a KeyError.

    The constructor does not accept key/value pairs to be created
    as dictionary members - pass a plain dictionary as first argument
    should be passed if that is needed.

    The keyword arg "key" to the constructor can take a callable analog to the
    "key" argument to "sorted" and "list.sort": it will
    accept one parameter and will be passed the insert/search
    key on all insertions and retrievals. Thus the internal
    order of the underlying tree can be fully customized.

    Having an inner AVL tree and a custom key function,  extra
    features are that `__getitem__` can retrieve a range of itens,
    and `get_closest_keys` will return a tuple of surrounding keys that exist.

    """

    node_cls = AVLNode

    def __init__(self, *args, key=None):
        self.key = key
        self.root = EmptyNode
        if len(args) == 1 and isinstance(args[0], Mapping):
            args = args[0].items()
        for key, value in args:
            self[key] = value

    def __getitem__(self, key):
        if not self.root:
            raise KeyError(key)
        if isinstance(key, slice):
            return [n.value for n in self.root.iter_slice(key)]
        return self.root.get(key).value

    def __setitem__(self, key, value):
        if self.root in (None, EmptyNode):
            self.root = self.node_cls(key=key, value=value, key_func=self.key)
        else:
            self.root.insert(key=key, value=value)

    def __delitem__(self, key):
        if not self.root:
            raise KeyError(key)
        if len(self.root) == 1:
            self.root = EmptyNode
        else:
            self.root.delete(key)

    def get_closest_keys(self, key):
        if not self.root:
            return None, None
        parent1, parent2 = self.root.get_closest(key)
        return (parent1.key if parent1 else None), (parent2.key if parent2 else None)

    def __iter__(self):
        return (n.key for n in self.root) if self.root else iter(())

    def __len__(self):
        return len(self.root) if self.root else 0

    def __repr__(self):
        return f"{self.__class__.__name__}({', '.join('%r=%r' % (k, v) for k, v in self.items())}{', key_func= %r' % (self.key) if self.key else ''})"



import pytest

from extradict.trie import PrefixCharTrie, PatternCharTrie, NormalizedTrie
from extradict.trie import _WORD_END, _ENTRY_END


def test_chartrie_can_store_text():
    a = PrefixCharTrie()
    a.add("test")
    assert "test" in a


def test_chartrie_simple_prefix_works():
    a = PrefixCharTrie(initial=["car", "carpet"])
    assert a["car"].contents == {"car", "carpet"}


@pytest.mark.parametrize("cls", (PrefixCharTrie, PatternCharTrie))
def test_chartrie_len_works(cls):
    a = cls(initial=["car", "carpet"])
    assert len(a) == 2


@pytest.mark.parametrize("cls", (PrefixCharTrie, PatternCharTrie))
def test_chartrie_iter_works(cls):
    a = cls(initial=["car", "carpet"])
    assert sorted(a) == ["car", "carpet"]


@pytest.mark.parametrize("cls", (PrefixCharTrie, PatternCharTrie))
def test_chartrie_works_and_exclude_others(cls):
    a = cls(initial=["car", "carpet", "java", "javascript"])
    assert a["car"].contents == {"car", "carpet"}


@pytest.mark.parametrize("cls", (PrefixCharTrie, PatternCharTrie))
def test_chartrie_update_works(cls):
    a = cls(initial=["car", "carpet"])
    a.update(["java", "javascript"])
    assert a["jav"].contents == {"java", "javascript"}


@pytest.mark.parametrize("cls", (PrefixCharTrie, PatternCharTrie))
def test_chartrie_returns_empty_on_non_existing_pattern(cls):
    a = cls(initial=["car", "carpet"])
    assert len(a["java"]) == 0


@pytest.mark.parametrize("cls", (PrefixCharTrie, PatternCharTrie))
@pytest.mark.parametrize("sentinel", [_WORD_END, _ENTRY_END])
def test_trie_keys_cant_contain_sentinel_values(cls, sentinel):
    a = cls(initial=["car", "carpet"])
    with pytest.raises(ValueError):
        a.add("java" + sentinel)


def test_modifying_prefixed_trie_updates_original():
    a = PrefixCharTrie(initial=["car"])
    assert len(a) == 1
    a["car"].add("pet")
    assert len(a) == 2
    assert "carpet" in a
    assert "pet" not in a
    assert "carpet" in a["car"]


def test_modifying_prefixed_pattern_trie_raises_value_error():
    a = PatternCharTrie(initial=["car"])
    with pytest.raises(ValueError):
        a["car"].add("pet")


@pytest.mark.parametrize("cls", (PrefixCharTrie, PatternCharTrie))
def test_shallow_copy_creates_indepent_chartrie(cls):
    from copy import copy

    a = cls(initial=["car"])
    b = copy(a)
    b.add("carpet")
    assert len(b) == 2
    assert len(a) == 1
    assert "carpet" not in a


@pytest.mark.parametrize("cls", (PrefixCharTrie, PatternCharTrie))
def test_copy_method_creates_indepent_chartrie(cls):
    a = cls(initial=["car"])
    b = a.copy()
    b.add("carpet")
    assert len(b) == 2
    assert len(a) == 1
    assert "carpet" not in a


def test_discard_method_prefixtrie():
    a = PrefixCharTrie(initial=["car", "carpet"])
    a.discard("car")
    assert len(a) == 1
    assert "car" not in a and "carpet" in a


def test_discard_method_patterntrie_single_element():
    a = PatternCharTrie(initial=["car"])
    a.discard("car")
    assert len(a) == 0
    assert "car" not in a
    assert len(a["ar"]) == 0
    assert len(a["car"]) == 0


def test_discard_method_patterntrie_element_postfix():
    a = PatternCharTrie(initial=["car", "carpet"])
    a.discard("car")
    assert len(a) == 1
    assert "car" not in a
    assert "carpet" in a
    assert len(a["ar"]) == 1


def test_discard_method_patterntrie_element_prefix():
    a = PatternCharTrie(initial=["car", "carpet"])
    a.discard("carpet")
    assert len(a) == 1
    assert "car" in a
    assert "carpet" not in a
    assert len(a["ar"]) == 1


def test_discard_method_patterntrie_element_infix():
    a = PatternCharTrie(initial=["discar", "car", "carpet"])
    a.discard("car")
    assert len(a) == 2
    assert "car" not in a
    assert "carpet" in a
    assert "discar" in a
    assert len(a["ar"]) == 2


def test_pattern_chartrie_simple_prefix_works():
    a = PatternCharTrie(initial=["car", "carpet", "oscar"])
    assert a["car"].contents == {"car", "carpet", "oscar"}


def test_pattern_chartrie_contains_works():
    a = PatternCharTrie(initial=["car", "carpet", "oscar"])
    assert "car" in a
    assert "carpet" in a
    assert "oscar" in a
    assert "ca" not in a
    assert "scar" not in a
    assert "arpe" not in a


def test_normalized_trie_works():
    a = NormalizedTrie(initial=["maçã", "maca"])
    assert len(a) == 2
    assert len(a["mac"]) == 2
    assert len(a["maç"]) == 2
    assert "maca" in a
    assert "maçã" in a


def test_normalized_trie_discard():
    a = NormalizedTrie(initial=["maçã", "maca"])
    a.discard("maçã")
    assert "maca" in a
    assert "maçã" not in a
    assert len(a) == 1
    a.discard("maca")
    assert len(a) == 0

    a = NormalizedTrie(initial=["maçã", "maca"])
    a.discard("maca")
    assert "maca" not in a
    assert "maçã" in a
    assert len(a) == 1
    a.discard("maçã")
    assert len(a) == 0



from collections import UserDict
from collections.abc import MutableSet, Iterable
from threading import RLock

import math

_OVERHEAD = 3
_SIZE_MASK = 0x7F
_IN_USE_MASK = 0x80


class _BlobSliceSet(MutableSet):
    def __init__(self, parent, offset, length):
        self.parent = parent
        self.offset = offset
        self.length = length
        raw_data = self.parent.data[offset + 1 : offset + length + 1].split(
            b"\x00\x00", 1
        )[0]
        self.data = {word.decode("utf-8") for word in raw_data.split(b"\x00")}

    def __contains__(self, item):
        return item in self.data

    def __iter__(self):
        return iter(self.data)

    def __len__(self):
        return len(self.data)

    def add(self, item):
        self.data.add(item)
        self.offset = self.parent._store(self.offset, self.data)

    def discard(self, item):
        self.data.discard(item)
        self.parent._force_content(self.offset, self.parent._encode(self.data))

    def __repr__(self):
        return f"{self.__class__.__name__}({{{', '.join(self)}}})"


class _BlobSets:
    def __init__(self, size_hint=1024):
        """
        Binary map per entry: first byte lower 7 bits indicates the log(2, entry_size). (entry is 2** (value & 0x7f) long). high bit indicates whether entry is in use or is deleted.
        """
        self.data = bytearray(b"\x00" * size_hint)
        self.offsets = list()  # maybe just track "last_offset"?
        self.lock = RLock()

    def _encode(self, data):
        words = set()
        for word in data:
            if "\x00" in word:
                raise ValueError(
                    f"{self.__class__.__name__} can't contain words with a \\x00 xharacter"
                )
            words.add(word)
        return b"\x00".join(word.encode("utf-8") for word in words)

    def add_new(self, data: "Iterable[str]"):
        """Adds a new string-set in a newly allocated point in the blob. Returns data offset to be used in .get"""
        if isinstance(data, str):
            raise TypeError("data must be a  non-string iterable of strings")
        with self.lock:

            self.offsets.append(offset := self._get_next_offset())

            data_bytes = self._encode(data)

            size = len(data_bytes)
            log_req_size = self._get_log_size(size)

            self._ensure_size(offset + 2**log_req_size)

            self.data[offset] = _IN_USE_MASK | log_req_size
            self._force_content(offset, data_bytes)
        return offset

    def _store(self, offset, data):
        data_bytes = self._encode(data)
        size = len(data_bytes)
        capacity = 2 ** (self.data[offset] & _SIZE_MASK)
        if size + _OVERHEAD <= capacity:
            self._force_content(offset, data_bytes)
            return offset
        new_offset = self._realloc(offset, len(data_bytes) + _OVERHEAD)
        self._force_content(new_offset, data_bytes)
        return new_offset

    def _get_next_offset(self):
        if not self.offsets:
            offset = 0
        else:
            last_offset = self.offsets[-1]
            offset = last_offset + 2 ** (self.data[last_offset] & _SIZE_MASK)
        return offset

    def _get_log_size(self, byte_size):
        # The "_OVERHEAD" is to ensure space for 1 header byte + 2 byte termination sequence (b"\x00\x00")
        return math.ceil(
            max(math.log(byte_size + _OVERHEAD, 2), 3)
        )  # minimal 8 bytes: allow posting a forwarding offset

    def _ensure_size(self, req_size):
        with self.lock:
            if len(self.data) < req_size:
                # ad-hoc heuristics follows: double size until we reach 4MB then go 4MB at a time
                increase = min(2**22, len(self.data))
                self.data.extend(b"\x00" * increase)

    def _realloc(self, old_offset, new_size):
        with self.lock:
            self._ensure_size(new_size + self.offsets[-1] if self.offsets else 0)
            content = self.data[
                old_offset + 1 : old_offset + 2 ** (self.data[old_offset] & _SIZE_MASK)
            ]
            self.data[old_offset] &= _SIZE_MASK
            self.offsets.append(new_offset := self._get_next_offset())
            self.data[new_offset] = _IN_USE_MASK | self._get_log_size(new_size)
            self.data[new_offset + 1 : len(content)] = content
            self.data[old_offset + 1 : old_offset + 7] = new_offset.to_bytes(
                6, "little"
            )
        return new_offset

    def _final_offset(self, offset):
        "Tracks the final destination of existing data"
        while not (self.data[offset] & _IN_USE_MASK):
            offset = int.from_bytes(self.data[offset + 1 : offset + 7], "little")
        return offset

    def _force_content(self, offset, data_bytes):
        with self.lock:
            self.data[offset + 1 : offset + len(data_bytes) + _OVERHEAD] = (
                data_bytes + b"\x00\x00"
            )

    def get(self, offset):
        """Retrieves the data contents at the given offset as a set-like object

        Automatically picks the new offset if the orignal data has been realocated
        at some point.

        (as of Python 3.12 an ampty set is 216 bytes long on x86_64 -
        we are down to 16 or 20 bytes per entry with up to 3 single characters here)
        """
        offset = self._final_offset(offset)
        return _BlobSliceSet(self, offset, 2 ** (self.data[offset] & _SIZE_MASK))


class BlobTextDict(UserDict):
    """Mapping wrapper over _BlobSets

    A dictionary with arbitrary keys and sets of string as values
    which will hold data in a single binary blob, instead of one
    Python set instance for each entry.
    """

    def __init__(self, *args, **kwargs):
        self.blob = _BlobSets()
        super().__init__(*args, **kwargs)

    def __setitem__(self, key, value: "Iterable[str]"):
        value_offset = self.blob.add_new(value)
        super().__setitem__(key, value_offset)

    def __getitem__(self, key):
        try:
            offset = super().__getitem__(key)
        except KeyError:
            super().__setitem__(key, offset := self.blob.add_new(set()))

        return self.blob.get(offset)

    def __repr__(self):
        return f"{self.__class__.__name__}({{{', '.join(key + ':' + repr(value) for key, value in self.items())}}})"



import sys

from collections.abc import Mapping


def namedtuple(name, attrs):
    """
    Alternate, clean room, implementation of 'namedtuple' as in stdlib's collection.namedtuple
    . This does not make use of "eval" at runtime - and can be up to 10 times faster to create
    a namedtuple class than the stdlib version.

    Instead, it relies on closures to do its magic.

    However, these will be slower to instantiate tahn stdlib version. The "fastnamedtuple"
    is faster in all respects, although it holds the same API for instantiating as tuples, and
    performs no lenght checking.

    """
    if isinstance(attrs, str):
        attrs = attrs.split()
    attrs = tuple(attrs)

    _field_order = {field: i for i, field in enumerate(attrs)}

    class NamedTuple(tuple):
        __slots__ = ()

        def __getattribute__(self, attr):
            n = _field_order.get(attr, None)
            if n is not None:
                return self[n]
            return tuple.__getattribute__(self, attr)

        def __new__(cls, *args, **kw):
            if len(args) + len(kw) != len(attrs):
                # to replicate the exact collections.namedtuple error messages here  mean to duplicate
                # Python's argument parsing logic itself. A meaningful TypeError should suffice
                # raise TypeError("__new__() missing {} required positional argument{}: {}")
                raise TypeError(
                    "__new__() got an incorrect number of parameters for '{}'".format(
                        name
                    )
                )
            if kw:
                original = args
                args = list(args) + [None] * len(kw)
                for key, value in kw.items():
                    pos = _field_order.get(key, None)
                    if pos is None:
                        raise TypeError(
                            "__new__() got an unexpected keyword argument '{}'".format(
                                key
                            )
                        )
                    if pos < len(original):
                        raise TypeError(
                            "__new__() got multiple values for argument '{}'".format(
                                key
                            )
                        )
                    args[pos] = value
            return tuple.__new__(cls, args)

        def __repr__(self):
            return "{}({})".format(
                self.__class__.__name__,
                ", ".join(
                    "{}={}".format(name, value)
                    for name, value in zip(self._fields, self)
                ),
            )

        def _asdict(self):
            return {key: value for key, value in zip(self._fields, self)}

        _fields = __definition_order__ = attrs

    NamedTuple.__name__ = name
    NamedTuple.__qualname__ = ".".join(
        (sys._getframe().f_back.f_globals.get("__name__", "__main__"), name)
    )

    return NamedTuple


def fastnamedtuple(name, attrs):
    """
    Like namedtuple but the class returned take an iterable for its values
    rather than positioned or named parameters. No checks are made towards the iterable
    lenght, which should match the number of attributes
    It is faster for instantiating as compared with stdlib's namedtuple
    """
    cls = namedtuple(name, attrs)
    delattr(cls, "__new__")
    cls.__qualname__ = ".".join(
        (sys._getframe().f_back.f_globals.get("__name__", "__main__"), name)
    )
    return cls


def defaultnamedtuple(_name, _attrs=None, **kw):
    """
    Implementation of named-tuple using default parameters -
    Either pass a sequence of 2-tupes or an ordered Mapping (like a dict) as the second parameter, or
    send in kwargs with the default parameters, after the first.
    (This takes advantadge of python  guaranteed ordering of **kwargs for a function
    see https://docs.python.org/3.6/whatsnew/3.6.html)

    """
    name = _name

    if _attrs:
        if not isinstance(_attrs, Mapping):
            _attrs = dict(_attrs)
        if kw:
            raise TypeError(
                "'defaultnamedtuple' should be passed either a dict or named parameters - not both"
            )
        kw = _attrs

    NamedTuple = namedtuple(name, kw.keys())

    class DefaultNamedTuple(NamedTuple):
        __slots__ = ()
        _defaults = kw

        def __new__(cls, *args, **kw):
            used = set(cls._fields[: len(args)])
            if kw:
                parameters = kw.keys()
                if used.intersection(parameters):
                    raise TypeError(
                        "__new__() got multiple values for arguments '{}'".format(
                            used.intersection(parameters)
                        )
                    )
                used.update(parameters)
            args += tuple(
                kw.get(attrname, cls._defaults[attrname])
                for attrname in cls._fields[len(args) :]
            )

            return tuple.__new__(cls, args)

    DefaultNamedTuple.__name__ = name
    DefaultNamedTuple.__qualname__ = ".".join(
        (sys._getframe().f_back.f_globals.get("__name__", "__main__"), name)
    )

    return DefaultNamedTuple



"""
Normalized Dict

In a key mismatch, fall back to key with normalized text:
just lower case letters and digits, with punctuation and whitespace removed.



"""

from collections.abc import MutableMapping
from threading import RLock

import re
import unicodedata


# Normalizer pipeline functions:

strip_replacer = lambda text: re.sub(r"\W", "", text)
unicode_normalizer = lambda text: unicodedata.normalize("NFKD", text)
case_normalizer = str.lower


class Normalizer:
    """Contains the standard normalizer method.

    It is used as a mixin for the NormalizedDict classes -
    the idea of having this is a class method is to make it possible
    to override the normalizer pipeline (or method) with a custom implementation through
    inheritance or simply  method attribution in the derived classes.

    """

    # Default pipeline decomposes any diactrics,
    # strips everything that is non alphanumeric
    # puts everything in lower case:
    pipeline = [unicode_normalizer, strip_replacer, case_normalizer]

    def normalize(self, text):
        for stage in self.pipeline:
            text = stage(text)
        return text


SENTINEL = object()


class _FallbackNormalizedDict(MutableMapping, Normalizer):
    """Dictionary meant for text only keys:
    will normalize keys in a way that capitalization, whitespace and
    punctuation will be ignored when retrieving items.

    A parallel dictionary is maintained with the original keys,
    so that strings that would clash on normalization can still
    be used as separated key/value pairs if original punctuation
    is passed in the key.

    Primary use case if for keeping translation strings when the source
    for the original strings is loose in terms of whitespace/punctuation
    (for example, in a html snippet)
    """

    def __init__(self, *args, **kw):
        self.literal = dict(*args, **kw)
        self.normalized = {
            self.normalize(key): value for key, value in self.literal.items()
        }

    def __getitem__(self, key):
        result = self.literal.get(key, SENTINEL)
        if result != SENTINEL:
            return result
        return self.normalized[self.normalize(key)]

    def __setitem__(self, key, value):
        self.literal[key] = value
        self.normalized[self.normalize(key)] = value

    def __delitem__(self, key):
        if key in self.literal:
            original_value = self.literal.pop(key)
        else:
            raise KeyError(key)
        norm_key = self.normalize(key)
        values = self.normalized[norm_key]
        values.remove(original_value)

    def __iter__(self):
        return self.literal.__iter__()

    def __len__(self):
        return len(self.literal)

    def normalized_keys(self):
        return self.normalized.keys()

    def __repr__(self):
        return "{}({})".format(self.__class__.__name__, repr(self.literal))


class FallbackNormalizedDict(MutableMapping, Normalizer):
    """Dictionary meant for text only keys:
    will normalize keys in a way that capitalization, whitespace and
    punctuation will be ignored when retrieving items.

    Multiple values set with keys that map to the same value
    are kept, and a list of all values set is returned
    on `__getitem__`.


    It is used internally by extradict.trie.NormalizedTrie so that
    similar words like "maçã" and "maca" can be part of the same
    trie, cohexist, and be both retrievable by the normalized form "maca"
    """

    def __init__(self, *args, **kw):
        self.lock = RLock()
        self.literal = dict(*args, **kw)
        self.normalized = {}
        self._update(self.literal)

    def _update(self, data):
        for key, value in data.items():
            self._setitem(key, value)

    def get_multi(self, key, default=SENTINEL):
        """Retrieves a list with all values associated the normalized form of the key"""
        return self.normalized.get(
            self.normalize(key), default if default is not SENTINEL else []
        )

    def __getitem__(self, key):
        result = self.literal.get(key, SENTINEL)
        if result != SENTINEL:
            return result
        try:
            return self.normalized[self.normalize(key)][0]
        except IndexError:
            raise KeyError(key)

    def __setitem__(self, key, value):
        prev = SENTINEL
        with self.lock:
            if key in self.literal:
                prev = self.literal.pop(key)
            self.literal[key] = value
            self._setitem(key, value, prev)
        return

    def _setitem(self, key, value, prev=SENTINEL):
        with self.lock:
            norm_key = self.normalize(key)
            if prev is not SENTINEL:
                index = self.normalized[norm_key].index(prev)
                self.normalized[norm_key][index] = value
            else:
                self.normalized.setdefault(norm_key, []).append(value)
        return

    def __delitem__(self, key):
        with self.lock:
            if key in self.literal:
                original_value = self.literal.pop(key)
            else:
                raise KeyError(
                    f"For deleting, the exact key used to set a value must be used. {key} not found"
                )
            norm_key = self.normalize(key)
            values = self.normalized[norm_key]
            values.remove(original_value)

    def __iter__(self):
        return self.literal.__iter__()

    def __len__(self):
        return len(self.literal)

    def normalized_keys(self):
        return self.normalized.keys()

    def __repr__(self):
        return "{}({})".format(self.__class__.__name__, repr(self.literal))


class NormalizedDict(MutableMapping, Normalizer):
    """Dictionary meant for text only keys:
    will normalize keys in a way that capitalization, whitespace and
    punctuation will be ignored when retrieving items.

    Unlike the FallbackNormalizedDict this does not keep the original
    version of the keys.

    """

    def __init__(self, *args, **kw):
        self.normalized = {
            self.normalize(key): value for key, value in dict(*args, **kw).items()
        }

    def __getitem__(self, key):
        return self.normalized[self.normalize(key)]

    def __setitem__(self, key, value):
        self.normalized[self.normalize(key)] = value

    def __delitem__(self, key):
        del self.normalized[self.normalize(key)]

    def __iter__(self):
        return self.normalized.__iter__()

    def __len__(self):
        return len(self.normalized)

    def __repr__(self):
        return "{}({})".format(self.__class__.__name__, repr(self.normalized))



from abc import ABC
from collections.abc import (
    MutableMapping,
    MutableSequence,
    Mapping,
    Sequence,
    Set,
    Iterable,
)
from copy import deepcopy
from textwrap import indent

import typing as T


_sentinel = object()
_strings = (str, bytes, bytearray)


class _NestedBase:
    @staticmethod
    def _get_next_component(key):
        if key is None:
            return None, None
        if isinstance(key, _strings):
            parts = key.split(".", 1)
        elif len(key) >= 2:
            parts = key[0], key[1:]
        else:
            parts = key
        if len(parts) == 1:
            return (parts[0] if parts[0] or parts[0] == 0 else None), None
        return parts

    @classmethod
    def wrap(cls, obj, clean=True):
        # clean:  ensures nested components are not wrapped.
        if clean:
            obj = cls.unwrap(obj, recurse=True)
        if isinstance(obj, __class__):
            return obj
        new_cls = _find_best_container([obj])
        if isinstance(new_cls, type) and issubclass(new_cls, __class__):
            self = new_cls()
            self.data = obj
            return self

        return obj

    @classmethod
    def unwrap(cls, obj, recurse=False):
        obj = obj if not isinstance(obj, __class__) else obj.data
        if (
            recurse
            and isinstance(obj, (Mapping, Sequence))
            and not isinstance(obj, _strings)
        ):
            items = obj.items() if isinstance(obj, Mapping) else enumerate(obj)
            for key, value in items:
                obj[key] = cls.unwrap(value, recurse)
        return obj

    def __eq__(self, other):
        return self.data == self.unwrap(other)

    def __len__(self):
        return len(self.data)

    def __repr__(self):
        def indent_content(r):
            if "\n" not in r:
                return r
            offset = len(r) - len(r.lstrip("[{"))
            r = (
                r[:offset]
                + ("\n" if r[offset] != "\n" else "")
                + indent(r[offset:], "    ")
            )
            return r

        if isinstance(self, Mapping):
            key_repr = []
            for k, v in self.items():
                r = f"{repr(k)}: "
                if isinstance(v, __class__):
                    r += indent_content(repr(v))
                else:
                    r += f"<{type(v).__name__}>"
                key_repr.append(r)
            return "{{{}}}".format(",\n".join(key_repr))

        else:  # Sequence
            if not len(self):
                return "[]"
            if isinstance(self[0], __class__):
                return "[{}]".format(
                    indent_content(repr(self[0]))
                    + (f"\n...\nx {len(self)}\n" if len(self) > 1 else "")
                )
            else:
                return repr(list(self))


class _NestedDict(_NestedBase, MutableMapping):
    def __init__(self, *args, **kw):
        self.data = {}

        if args and kw:
            raise TypeError(
                f"{self.__class__} should either get positional arguments or named arguments"
            )
        if len(args) == 1 or kw:
            initial_mapping = args[0] or kw
            items = initial_mapping.items()
        else:
            items = args

        for key, value in items:
            self[key] = value

    def __getitem__(self, key):
        key, subpath = self._get_next_component(key)
        # TBD: upon "key" being a path element that would select multiple components,
        # build and return a NestedDataQuery object

        if subpath:
            return self[key][subpath]
        value = self.data[key]
        return self.wrap(value)

    def merge(self, data, path=""):
        self._setitem(path, self.unwrap(data), merging=True)
        return self

    def __setitem__(self, key: str, value: T.Any):
        return self._setitem(key, self.unwrap(value), merging=False)

    def _setitem(self, key: str, value: T.Any, merging: bool):
        # cyclomatic complexity is my bitch
        if isinstance(value, (_NestedDict, _NestedList)):
            value = value.data

        key, subpath = self._get_next_component(key)

        if not key:  # we are modifying the key contents at the root of this object
            if not isinstance(value, Mapping):
                raise TypeError(
                    f"{self.__class__.__name__} root value must be set to a mapping"
                )
            if not merging:
                self.data = deepcopy(value)
                return

            for new_key, new_value in value.items():
                if new_key in self.data:
                    sub_item = self[new_key]
                    if isinstance(sub_item, NestedData):
                        sub_item.merge(new_value, "")
                        continue
                self[new_key] = deepcopy(new_value)
            return

        if key not in self.data:
            if subpath:
                self.data[key] = {}
                self[key]._setitem(subpath, value, merging)
            else:
                if isinstance(value, Mapping) and _should_be_a_sequence(value):
                    value = _extract_sequence(value)
                self.data[key] = value
            return
        if subpath:
            if not isinstance(self.data[key], (Mapping, Sequence)):
                next_key, next_subpath = self._get_next_component(subpath)
                if next_key.isdigit():
                    if next_key == "0":
                        self[key] = [{}]
                    else:
                        raise IndexError("New sequence not starting at 0 when merging")
                else:
                    self[key] = {}
            if not merging:
                self[key][subpath] = value
                return
            if isinstance(self[key], Sequence):
                self[key].merge(value, subpath)
                return
            self[key]._setitem(subpath, value, merging)
            return

        if isinstance(self.data[key], Mapping):
            if isinstance(value, Mapping):
                subitem = self[key]
                for sub_key, sub_value in value.items():
                    subitem._setitem(sub_key, sub_value, merging)
        elif isinstance(self.data[key], Sequence) and not isinstance(
            self.data[key], _strings
        ):
            if isinstance(value, Mapping) and key in value:
                value = value[key]
            self[key].merge(value)
        else:
            self.data[key] = value

    def __delitem__(self, key):
        if key in self.data:
            del self.data[key]
            return
        key, subpath = self._get_next_component(key)
        if key not in self.data:
            raise KeyError(key)
        del self[key][subpath]

    def __iter__(self):
        return iter(self.data)

    def __contains__(self, key):
        key, subpath = self._get_next_component(key)
        if key not in self.data:
            return False
        if not subpath:
            return True
        return subpath in self[key]

    def walk(self):
        pass

    def __len__(self):
        return len(self.data)


def _extract_sequence(obj, default=_sentinel):
    elements = []
    if not obj:
        # Normally not reachable: an empty mapping should be created as a mapping
        return elements
    target = int(max(obj.keys(), key=int))
    for i in range(target + 1):
        element = obj.get(i, obj.get(str(i), _sentinel))
        # the sentiel usage in the next lines is completly decoupled:
        if element is _sentinel:
            if default is not _sentinel:
                element = default if not callable(default) else default()
            else:
                raise ValueError(f"Missing index {i} on NestedData sequence")
        elements.append(element)
    return elements


class _NestedList(_NestedBase, MutableSequence):
    def __init__(self, *args, default=_sentinel):
        if len(args) == 1 and isinstance(args[0], Iterable):
            args = args[0]
        if isinstance(args, Mapping):
            args = _extract_sequence(args, default=default)
        else:
            # even if it is a list, we make a shallow copy:
            args = list(args)
        self.data = args

    def __getitem__(self, index):
        # No sense in accepting "*": just take the whole list.
        if isinstance(index, slice):
            return self.__class__(self.data[index])
        if not isinstance(index, int):
            index, subpath = self._get_next_component(index)
        else:
            subpath = None
        if index == "*":
            result = [(comp[subpath] if subpath else comp) for comp in self]
            return self.wrap(result)
        wrapped = self.wrap(self.data[int(index)])
        if subpath:
            return wrapped[subpath]
        return wrapped

    def __setitem__(self, index, item):
        if isinstance(index, slice):
            self.data[index] = item
            return
        if not isinstance(index, int):
            index, subpath = self._get_next_component(index)
        else:
            subpath = None
        if index == "*":
            for i, element in enumerate(self):
                if subpath:
                    element[subpath] = item
                else:
                    self[i] = item
            return
        if subpath is None:
            self.data[int(index)] = (
                item if not isinstance(item, NestedData) else item.data
            )
        else:
            self.wrap(self[index])[subpath] = item

    def __delitem__(self, index):
        if isinstance(index, slice):
            del self.data[index]
            return
        if not isinstance(index, int):
            index, subpath = self._get_next_component(index)
        else:
            subpath = None
        if subpath is None:
            del self.data[int(index)]
        del self.wrap(self[index])[subpath]

    def insert(self, index, item):
        try:
            index = index.__index__()
        except Exception as error:
            raise TypeError from error
        self.data.insert(
            int(index), item if not isinstance(item, NestedData) else item.data
        )

    def merge(self, data, path=None):
        data = self.unwrap(data)
        if path in (None, ""):
            if isinstance(data, Sequence) and not isinstance(data, _strings):
                if len(self) != len(data):
                    if len(data) == 1:
                        data = [deepcopy(data[0]) for _ in range(len(self))]
                    else:
                        raise ValueError(
                            f"Sequences to merge are not same size. len(self) == {len(self)}, len(data) == {len(data)}"
                        )
                for item, data_item in zip(self, data):
                    if isinstance(item, NestedData):
                        item.merge(data_item)
                return self
            raise ValueError("Trying to merge a mapping into a sequence")
        elif isinstance(path, str):
            key, subpath = self._get_next_component(path)
            if not key:
                raise IndexError("No index to perform merging on")
        else:
            key = str(path.__index__())  # let it burn if this fails.
            subpath = ""
        if key == "*":
            item_iterable = enumerate(self)
        else:
            item_iterable = [(key, self[key])]
        for i, item in item_iterable:
            if isinstance(item, NestedData):
                item.merge(data, subpath)
            elif not subpath:
                self[i] = data
            else:
                raise IndexError(
                    "Incompatible item with further merging: item {item!r} at position {i} is not a container which could fit {subpath}"
                )

        return self


def _should_be_a_sequence(obj, default=_sentinel, **kw):
    if not obj:
        return False
    if isinstance(obj, Set):
        return True
    if all(
        isinstance(k, int) or (isinstance(k, _strings) and k.isdigit())
        for k in obj.keys()
    ):
        if default:
            return True
        if len(set(map(int, obj.keys()))) == len(obj) and 0 in obj or "0" in obj:
            return True
    return False


def _find_best_container(args, kw=None):
    if kw is None:
        kw = {}
    if len(args) == 1 and (
        not isinstance(args[0], (Sequence, Mapping, Set))
        or isinstance(args[0], _strings)
    ):
        return lambda x: x
    if len(args) == 0 and kw and any(k for k in kw.keys() if not k.isdigit()):
        return _NestedDict
    if len(args) == 1:
        if isinstance(args[0], Mapping):
            return _NestedList if _should_be_a_sequence(args[0]) else _NestedDict
        return _NestedList
    if len(args) > 1:
        return _NestedList
    return _NestedDict


class NestedData(ABC):
    """Nestable mappings and sequences data structure to facilitate field access

    The idea is a single data structure that can hold "JSON" data, adding some
    helper methods and functionalities.

    Primarily, one can use a dotted string path to access a deply nested key, value pair,
    instead of concatenating several dictionary ".get" calls.

    Examples:
         person["address.city"] instead of person["address"]["city"]

         or

         persons["10.contacts.emails.0"]


    The first tool available is the ability to merge mappings with extra keys
    into existing nested mappings, without deleting non colidng keys:
    a "person.address" key that would contain "city" but no "street" or "zip-code"
    can be updated with:  `record["person"].merge({"address": {"street": "5th ave", "zip-code": "000000"}})`
    preserving the "person.address.city" value in the process.

    The ".data" attribute stores the object contents as a tree of dicionary and lists as needed -
    these are lazily wrapped as NestedData instances if retrieved through the class, but
    can be freely manipulated directly.


    """

    # implemented as a class so that structures can be tested as instances of it.
    # actually this skelleton just a dispatcher factory function, that works
    # as "virtual parent" to the real data structures
    def __new__(cls, *args, **kw):

        cls = _find_best_container(args, kw)

        return cls(*args, **kw)

    def merge(self, data, path):
        """Will insert or change new keys present in data into the existing data structure,
        starting at "path".
        """
        pass


# Virtual Subclassing so that both "_NestedDict" and "_NestedList"s show up as instances of "NestedData"
NestedData.register(_NestedDict)
NestedData.register(_NestedList)



# coding:utf-8
from .version_dict import VersionDict
from .version_dict import OrderedVersionDict
from .normalized_dict import FallbackNormalizedDict
from .normalized_dict import NormalizedDict
from .map_getter import MapGetter
from .reciprocal_dict import BijectiveDict
from .extratuple import namedtuple
from .extratuple import defaultnamedtuple
from .extratuple import fastnamedtuple
from .binary_tree_dict import TreeDict
from .grouper import Grouper
from .nested_data import NestedData
from .trie import PrefixTrie, Trie, NormalizedTrie
from .blobdict import BlobTextDict

__author__ = "João S. O. Bueno"
__version__ = "0.7.0beta1"


__all__ = [
    "VersionDict",
    "OrderedVersionDict",
    "FallbackNormalizedDict",
    "NormalizedDict",
    "MapGetter",
    "BijectiveDict",
    "namedtuple",
    "defaultnamedtuple",
    "fastnamedtuple",
    "TreeDict",
    "Grouper",
    "NestedData",
    "PrefixTrie",
    "Trie",
    "NormalizedTrie",
    "BlobTextDict",
]



from collections.abc import MutableSet
from copy import copy
from threading import RLock

from .blobdict import BlobTextDict
from .normalized_dict import FallbackNormalizedDict as NormalizedDict

# These two are officially stated by Unicode as "not characters".
# will do fine as sentinels:
_WORD_END = "\ufffe"
_ENTRY_END = "\uffff"


class PrefixCharTrie(MutableSet):
    """A prefix-based Trie for strings with a Set interface.

    use "CharTrie[prefix].contents" to retrieve a set of all strings with the given prefix

    CharTrie[prefix] will return another instance of CharTrie **that shares** the same underlying
    data to the parent trie: additions or deletions will reflect on the parent and other siblings.

    Use `CharTrie[prefix].copy()` to have an independent data structure.
    """

    def __init__(self, initial=None, *, root=None, pattern="", backend=None):
        if backend is None:
            backend = dict
        self.backend = backend
        self.data = root if root is not None else backend()
        self.pattern = pattern
        self.lock = RLock()
        if backend is not BlobTextDict and not pattern in self.data:
            self.data[pattern] = set()

        if initial:
            self.update(initial)

    def __getitem__(self, key):
        pattern = self.pattern
        for letter in key:
            if letter in self.data[pattern]:
                pattern += letter
                continue
            return self.__class__(backend=self.backend)
        with self.lock:
            sub_instance = self._clone()
        sub_instance.pattern = pattern
        return sub_instance

    def _clone(self):
        new = self.__class__.__new__(self.__class__)
        new.backend = self.backend
        new.data = self.data
        new.lock = self.lock
        new.pattern = self.pattern
        return new

    @property
    def contents(self):
        return self._contents(self.pattern)

    def _contents(self, pattern):
        results = set()
        if pattern not in self.data:
            return results
        for item in self.data[pattern]:
            if item == _WORD_END:  # It is not because it is used as a singleton around
                # that one can use "is": once merged to a string and sliced
                # from there you have new instances. (thanks unit tests!)
                results.add(pattern)
                continue
            results.update(subitem for subitem in self._contents(pattern + item))
        return results

    def add(self, key):
        if _ENTRY_END in key or _WORD_END in key:
            raise ValueError("Invalid character in key")
        key = key + _WORD_END
        with self.lock:
            pattern = self.pattern
            for letter in key:
                if self.backend is not BlobTextDict:
                    branch = self.data.setdefault(pattern, set())
                else:
                    branch = self.data[pattern]
                pattern = pattern + letter
                branch.add(letter)
            self.data[pattern] = _WORD_END

    def copy(self):
        cls = type(self)
        return cls(self.contents)

    def __copy__(self):
        new_instance = type(self)(pattern=self.pattern, root=self.data.copy())
        return new_instance

    def __contains__(self, key):
        return self.data.get(key + _WORD_END, None) == _WORD_END

    def discard(self, key):
        # FIXME: this removes the complete key, but various "unfinshed prefixes"
        # that would lead to the discarded key remain in self.data
        if key + _WORD_END not in self.data:
            raise KeyError()
        del self.data[key + _WORD_END]
        self.data[key].remove(_WORD_END)

    def update(self, seq):
        for item in seq:
            self.add(item)

    def __iter__(self):
        yield from iter(self.contents)

    def __len__(self):
        return len(self.contents)

    def __repr__(self):
        return f"Trie {('prefixed with ' + repr(self.pattern)) if self.pattern else ''} with {len(self)} elements."


class PatternCharTrie(PrefixCharTrie):
    """A pattern-based Trie for strings with a Set interface.

    use "PatternCharTrie[prefix].contents" to retrieve a set of all strings containing the given pattern

    PatternCharTrie[prefix] will return another trie instance **that shares** the same underlying
    data to the parent trie: additions or deletions will reflect on the parent and other siblings.

    Use `PatternCharTrie[prefix].copy()` to have an independent data structure.
    """

    def _subpattern_add(self, key):
        if len(key) < 2:
            return
        pattern = ""
        for char in key:
            branch = self.data.setdefault(pattern, set())
            pattern += char
            branch.add(char)

    def add(self, key):
        if _ENTRY_END in key or _WORD_END in key:
            raise ValueError("Invalid character in key")

        if self.pattern:
            raise ValueError(
                "PatternCharTrie cannot add new final values having a selected pattern"
            )

        with self.lock:
            for i in range(len(key)):
                self._subpattern_add(
                    key[i:] + (_WORD_END + key[:i] if i else "") + _ENTRY_END
                )

    def _demangle(self, word):
        if not _WORD_END in word:
            return word
        ending, start = word.split(_WORD_END)
        return start + ending

    def _contents(self, pattern):
        results = set()
        if pattern not in self.data:
            return results
        for item in self.data[pattern]:
            if item == _ENTRY_END:
                results.add(self._demangle(pattern))
                continue
            results.update(
                subitem for subitem in __class__._contents(self, pattern + item)
            )
        return results

    def discard(self, key):
        if key not in self and not self._deleting_guard(key):
            raise KeyError(f"No item corresponding to {key}")
        with self.lock:
            for i, letter in enumerate(key):
                # root = self.data[""]
                partial = letter
                paths = {
                    partial,
                }
                for j, next_letter in enumerate(
                    key[i + 1 :] + (_WORD_END if i > 0 else "") + key[0:i]
                ):
                    partial += next_letter
                    paths.add(partial)
                self.data[partial].remove(_ENTRY_END)

                # trying to remove the "fossils" that led to the
                # removed entry is very hard, if these
                # are used by more words.
                # we just leave them there!

    def __contains__(self, key):
        return _ENTRY_END in self.data.get(key, set())

    def _deleting_guard(self, key):
        return False

    def __repr__(self):
        return f"PatternTrie {('patterned with ' + repr(self.pattern)) if self.pattern else ''} with {len(self)} elements."


class NormalizedTrie(PatternCharTrie):
    def __init__(self, *args, **kwargs):
        self.normalized = NormalizedDict()
        super().__init__(*args, **kwargs)

    def __contains__(self, key):
        return key in self.normalized.literal

    def add(self, key):
        self.normalized[key] = key
        new_key = self.normalized.normalize(key)
        super().add(new_key)

    def __copy__(self):
        new = super().__copy__()
        new.normalized = copy(self.normalized)
        return new

    def _clone(self):
        new = super().__copy__()
        new.normalized = copy(self.normalized)
        return new

    def __getitem__(self, key):
        if key in self.normalized.literal:
            return {key}
        return super().__getitem__(self.normalized.normalize(key))

    def _contents(self, pattern):
        pattern = self.normalized.normalize(pattern)
        normalized_results = super()._contents(pattern)
        results = set()
        for norm_key in normalized_results:
            results.update(self._expand(norm_key))
        return results

    def _expand(self, pattern):
        return self.normalized.get_multi(pattern)

    def discard(self, key):
        with self.lock:
            del self.normalized[key]
            norm_key = self.normalized.normalize(key)
            if not self.normalized.get_multi(norm_key):
                self._deleting_key = norm_key
                super().discard(norm_key)
                del self._deleting_key

    def _deleting_guard(self, key):
        return getattr(self, "_deleting_key", None) == key

    @property
    def contents(self):
        return {self.normalized[item] for item in super().contents}


PrefixTrie = PrefixCharTrie
Trie = PatternCharTrie

__all__ = ["PrefixCharTrie", "Trie", "NormalizedTrie"]
